/*****************************************************************************
* UNIT DESCRIPTION:  Base class for input values and methods to load time series
*                    data from a file
* SPEC REF: (if used)
* DESIGN REF: (if used)
*
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************/
using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.IO;
using System.Diagnostics;
using SELGlobalStringMethods;

namespace SELChartClasses
{

    public enum THMIChannelPhase {cpA=0, cpB, cpC, cpOther};
    public enum THMIChannelType {ctCurrent=0, ctVoltage, ctPower, ctOther};
    public enum THMIEventPhase {epA=0, epB, epC, epAll};

    public enum TTimeSeriesDeviceType {dtSEL3XX, dtSEL4XX, dtSEL7XX};
    public enum TTimeSeriesColumnType {scRecord, scDate, scTime, scPhase, scStatus};
    public enum TTimeSeriesFileType {ftUnknown = -1, ftReport, ftCSV, ftText};
    public enum TTimeSeriesCBEMAViolationType {cvtNone = 0, cvt90, cvt80, cvt70, cvt0, cvt110, cvt120, cvt200};
    public enum TTimeSeriesEventType {evNone=-1, evSag=0, evSwell, evInterrupt, evAll, evMulti};
    public enum TTimeSeriesChartType {ctInvalid=-1,ctUndefined=0, ctSSI, ctLDP};
    public enum TTimeSeriesCompleteType {ctComplete=0, ctIncompleteStart, ctIncompleteStop, ctIncompleteBoth};

    public class TTimeSeriesSettingRecord 
    {
        private string FSettingName;
        private float FSettingValue;

        public string SettingName
        {
          get{return FSettingName;}
          set{FSettingName = value;}
        }

        public float SettingValue
        {
          get{return FSettingValue;}
          set{FSettingValue = value;}
        }
    }

    public struct TTimeSeriesRecord 
    {
        public int RecordNumber;
        public DateTime DateTime;
        public float[] Value;
        public string Phase;
        public string Status;
    }

    public struct TTimeSeriesIndexRecord
    {
        public DateTime DateTime;
        public int FirstRecord;    // first record of event
        public int LastRecord;    // last record of event
        public int StartEvent;    // end of pre-disturbance
        public int StopEvent;    // start of post-disturbance
        public string EventName;    // Event number
        public string Description;
        public TimeSpan Duration;   // span of time
        public TTimeSeriesEventType EventType;
        public double Magnitude;
        public double VAMin;
        public double VAMax;
        public double VBMin;
        public double VBMax;
        public double VCMin;
        public double VCMax;
        public Boolean Checked;    // user selected for display
        public TTimeSeriesCompleteType Completion;    // state of completion
        public TTimeSeriesCBEMAViolationType Violation;
    }

    public struct TTimeSeriesLDPEventRecord
    {
        public DateTime DateTime;
        public string EventName;
        public string EventCode;
        public string Description;
    }

    public struct TTimeSeriesLDPChannelRecord
    {
        public string ChannelName;
        public string ChannelFunction;
        public string ChannelUnits;
        public string ChannelRules;
        public float ChannelScale;
    }

    public class TTimeSeriesDataTable 
    {
        const int COL_RECORD = 0;
        const int COL_DATE = 1;
        const int COL_TIME = 2;
        const int COL_PHASE = 3;
        const int COL_STATUS = 4;

        const int RELAY_3XX_OFFSET = 1;   // one more column
        const int RELAY_3XX_START = 7;  // skip Vs
        const int RELAY_4XX_OFFSET = (-5);  // 5 less columns

        private List<string> FRawDataLabels = new List<string>(); // list of variable names
        private List<TTimeSeriesRecord> FRawDataTable = new List<TTimeSeriesRecord>();   // list of raw data records
        private List<TTimeSeriesIndexRecord> FIndexTable = new List<TTimeSeriesIndexRecord>();  // list of event index records
        private List<TTimeSeriesLDPEventRecord> FLDPEventTable = new List<TTimeSeriesLDPEventRecord>();  // list of LDP event records
        private List<TTimeSeriesLDPChannelRecord> FLDPChannelTable = new List<TTimeSeriesLDPChannelRecord>();  // list of LDP channel records
        private List<TTimeSeriesSettingRecord> FSettingTable = new List<TTimeSeriesSettingRecord>();  // List of setting names/values
        private TTimeSeriesInputSettings FInputSettings;   // input settings from GUI

        private string FChartName;  // Name of chart
        private TTimeSeriesChartType FChartType;  // Type of chart (based on name)
        private TTimeSeriesDeviceType FDeviceType;  // type of device (dt734XX)
        private string FChartVersion; // Chart version
        private DateTime FChartDateTime; // Chart Date/Time
        private string FChartField; // name of first column of charted table

        private string FMeterID;  // MID of meter
        private string FMeterFID;  // FID of meter
        private string FMeterTID;  // TID of meter
        private string FTimeSource;  // Time source

        private int[] FHeaderColumns; // list of header columns in charted table
        private int[] FDataColumns;   // list of data columns in charted table
        private string FDateFormat;   // Date format setting string (MDY,...)

        public List<string> RawDataLabels
        {
            get{return FRawDataLabels;}
            set{FRawDataLabels = value;}
        }

        public List<TTimeSeriesRecord> RawDataTable
        {
            get{return FRawDataTable;}
            set{FRawDataTable = value;}
        }

        public List<TTimeSeriesIndexRecord> IndexTable
        {
            get{return FIndexTable;}
            set{FIndexTable = value;}
        }

        public List<TTimeSeriesLDPEventRecord> LDPEventTable
        {
            get{return FLDPEventTable;}
            set{FLDPEventTable = value;}
        }

        public List<TTimeSeriesLDPChannelRecord> LDPChannelTable
        {
            get{return FLDPChannelTable;}
            set{FLDPChannelTable = value;}
        }

        public List<TTimeSeriesSettingRecord> SettingTable
        {
            get{return FSettingTable;}
            set{FSettingTable = value;}
        }

        public TTimeSeriesInputSettings InputSettings
        {
            get{return FInputSettings;}
            set{FInputSettings = value;}
        }

        public string ChartField
        {
            get{return FChartField;}
            set{FChartField = value;}
        }

        public string ChartName
        {
            get{return FChartName;}
            set{FChartName = value;}
        }

        public TTimeSeriesChartType ChartType
        {
            get{return FChartType;}
            set{FChartType = value;}
        }

        public TTimeSeriesDeviceType DeviceType
        {
            get{return FDeviceType;}
            set{FDeviceType = value;}
        }

        public string ChartVersion
        {
            get{return FChartVersion;}
            set{FChartVersion = value;}
        }

        public DateTime ChartDateTime
        {
            get{return FChartDateTime;}
            set{FChartDateTime = value;}
        }

        public string MeterID
        {
            get{return FMeterID;}
            set{FMeterID = value;}
        }

        public string MeterFID
        {
            get{return FMeterFID;}
            set{FMeterFID = value;}
        }

        public string MeterTID
        {
            get{return FMeterTID;}
            set{FMeterTID = value;}
        }

        public string TimeSource
        {
            get{return FTimeSource;}
            set{FTimeSource = value;}
        }

        public int[] DataColumnArray
        {
            get{return FDataColumns;}
            set{FDataColumns = value;}
        }

        public int[] HeaderColumnArray
        {
            get{return FHeaderColumns;}
            set{FHeaderColumns = value;}
        }

        public string DateFormat
        {
            get{return FDateFormat;}
            set{FDateFormat = value;}
        }
        
        public int[] HeaderColumns
        {
            get{return FHeaderColumns;}
            set{FHeaderColumns=value;}
        }
        public int[] DataColumns
        {
            get{return FDataColumns;}
            set{FDataColumns=value;}
        }


/******************************************************************************
* MODULE NAME ...  Create
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  TimeSeriesDataTable is used to hold channel data obtained
*                          from the device. The data may be used to generated a
*                          graphical chart or to export as CSV data file.
*
* INPUT PARAMETERS:
*                    None:
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Initial implementation
*******************************************************************************/
        public TTimeSeriesDataTable()
        {
            FChartField = "#";               // default column name for charted data
            FChartName = "None";
            FChartDateTime = DateTime.Now;
            FChartVersion = "1.0";
            FTimeSource = "Unknown";
            FMeterID = "SEL-734";  // MID of meter
            FMeterFID = "FID";  // FID of meter
            FMeterTID = "TID";  // TID of meter
            FDateFormat = "MM/dd/yyyy";
        }
/************************************************************************
* NAME:       Clear
* FUNCTIONAL DESCRIPTION:  Clears content of each table (without destroying the table)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void Clear()
        {
            FRawDataTable.Clear();
            FRawDataLabels.Clear();
            FIndexTable.Clear();
            FLDPEventTable.Clear();
            FLDPChannelTable.Clear();
            FSettingTable.Clear();
        }
/************************************************************************
* NAME:       SetLengthHeaderColumns
* FUNCTIONAL DESCRIPTION:    Set length of header columns
* INPUT PARAMETERS: size : integer array size
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void SetLengthHeaderColumns(int size)
        {
            FHeaderColumns = new int[size];
        }
/************************************************************************
* NAME:       SetLengthDataColumns
* FUNCTIONAL DESCRIPTION:    Set length of data columns
* INPUT PARAMETERS: size : integer array size
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void SetLengthDataColumns(int size)
        {
            FDataColumns = new int[size];
        }
/************************************************************************
* NAME:       GetHeaderColumn
* FUNCTIONAL DESCRIPTION:    Get a column from the header column array
* INPUT PARAMETERS: Index : integer column number
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public int GetHeaderColumn(int Index)
        {
          return FHeaderColumns[Index];
        }
/************************************************************************
* NAME:       SetHeaderColumn
* FUNCTIONAL DESCRIPTION:    Set a column from the data column array
* INPUT PARAMETERS: Index : integer  column number
*                   Value : integer  value
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void SetHeaderColumn(int Index, int Value)
        {
            FHeaderColumns[Index] = Value;
        }
/************************************************************************
* NAME:       GetDataColumn
* FUNCTIONAL DESCRIPTION:    Get a column from the data column array
* INPUT PARAMETERS: Index : integer  column number
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public int GetDataColumn(int Index)
        {
            return FDataColumns[Index];
        }
/************************************************************************
* NAME:       SetDataColumn
* FUNCTIONAL DESCRIPTION:    Set a column from the data column array
* INPUT PARAMETERS: Index : integer  column number
*                   Value : integer  value
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void SetDataColumn(int Index, int Value)
        {
            FDataColumns[Index] = Value;
        }
/******************************************************************************
* MODULE NAME ...  GetFileType
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:
* Determine file type and delimiter by examining record
*
* INPUT PARAMETERS:
*                    fileData: string                sample data record
*
* OUTPUT PARAMETERS:
*                    fileDelimiter: string   delimiter to use when parsing file
* RETURN VALUE:     fileType :  TTimeSeriesFileType  type of file (ftCSV or ftReport)
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.07  Kent Myers        Initial implementation
*******************************************************************************/
        public TTimeSeriesFileType GetFileType(List<string> fileData, ref char fileDelimiter)
        {
            TTimeSeriesFileType fileType;

            if(fileData.Count > 0) 
            {
                if(fileData[0].IndexOf(',') >= 0)  // it is a CSV file
                {
                  fileType = TTimeSeriesFileType.ftCSV;
                  fileDelimiter = ',';
                } else
                {
                  fileType = TTimeSeriesFileType.ftReport;
                  fileDelimiter = ' ';
                }
            }
            else
            {
                fileType = TTimeSeriesFileType.ftUnknown;
                fileDelimiter = ' ';
            }
            return fileType;
        }

/************************************************************************
* NAME:       ReadFileType
* FUNCTIONAL DESCRIPTION:    Read file type from file (if found)
* INPUT PARAMETERS: filePath : string  file name
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesFileType  type of file (ftCSV, ftReport)
************************************************************************/
    public TTimeSeriesFileType ReadFileType(string filePath)
    {
        List<string> fileData, recordData;
        TTimeSeriesFileType fileType;
        GlobalString gs = new GlobalString();
            
        fileType = TTimeSeriesFileType.ftUnknown;

        if(File.Exists(filePath)) 
        {
            fileData = new List<string>();
            recordData = new List<string>();
            try 
            {
                char delim = ' ';
             //fileData.LoadFromFile(filePath);
                gs.LoadFromFile(filePath,fileData);
                fileType = GetFileType(fileData,ref delim);
            }
            catch 
            {
              //fileData.Free;
              //recordData.Free;
            }
        }
        else
        {
            fileType = TTimeSeriesFileType.ftUnknown;
        }

        return fileType;
    }
/************************************************************************
* NAME:       ReadChartType
* FUNCTIONAL DESCRIPTION:    Read chart type from file (if found)
* INPUT PARAMETERS: filePath : string  file name
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public TTimeSeriesChartType ReadChartType(string filePath)
        {
            TTimeSeriesFileType fileType = TTimeSeriesFileType.ftUnknown;

            return ReadChartType(filePath,out fileType);
        }
/************************************************************************
* NAME:       ReadChartType
* FUNCTIONAL DESCRIPTION:    Read chart type from file (if found)
* INPUT PARAMETERS: filePath : string  file name
* OUTPUT PARAMETERS: fileType : TTimeSeriesFileType  type of file (ftReport,ftCSV)
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public TTimeSeriesChartType ReadChartType(string filePath, out TTimeSeriesFileType fileType)
        {
            const int RECORD_VERSION = 1;
            const int FIELD_CHARTNAME = 0;
            GlobalString gs = new GlobalString();

            List<string> fileData, recordData;
            char delim = ' ';
            TTimeSeriesChartType myChartType = TTimeSeriesChartType.ctUndefined;

            fileType = TTimeSeriesFileType.ftUnknown;

            if(File.Exists(filePath)) 
            {
                fileData = new List<string>();
                recordData = new List<string>();
                try
                {
                    gs.LoadFromFile(filePath,fileData);

                    fileType = GetFileType(fileData,ref delim);
                    if(fileType == TTimeSeriesFileType.ftUnknown) 
                    {
                        return TTimeSeriesChartType.ctUndefined;         // invalid file
                    }

                    //recordData.Delimiter = delim;

                    if((fileType == TTimeSeriesFileType.ftCSV) && (fileData.Count > RECORD_VERSION))    // get standard columns and data columns
                    {
                        gs.ParseString(delim,fileData[RECORD_VERSION],recordData);
                        //recordData.CommaText = fileData[RECORD_VERSION];    // get Chart record name
                        FChartName = recordData[FIELD_CHARTNAME];
                        myChartType = GetChartType(ChartName);
                    }
                    else if(fileType == TTimeSeriesFileType.ftReport) 
                    {
                        ChartField = "#";  // report file data delimiter
                        myChartType = GetReportType(fileData,recordData);
                    }
                    else
                    {
                        myChartType = TTimeSeriesChartType.ctUndefined;
                    }
                 }
                 catch
                 {
                 }
                //finally
                //  fileData.Free;
                //  recordData.Free;
                //end;
            }
            else
            {
                myChartType = TTimeSeriesChartType.ctUndefined;
            }
            return myChartType;
        }

/************************************************************************
* NAME:       GetChartType
* FUNCTIONAL DESCRIPTION:    Determine Chart type from chart name
* INPUT PARAMETERS:  name : string chart name
* OUTPUT PARAMETERS:  none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public TTimeSeriesChartType GetChartType(string name) 
        {
            string[] CHART_TYPES = {"None","SSI","VSSI","LDP"};

            int i,found;
            TTimeSeriesChartType myChartType;

            found = -1;
            for (i =0; i<CHART_TYPES.Length; i++) 
            {
                if(name == CHART_TYPES[i])
                {
                    found = i;
                    break;
                }
            }

            switch( found ) 
            {
                case 1:
                case 2: // SSI,VSSI
                    myChartType =  TTimeSeriesChartType.ctSSI;
                break;
                case 3:
                    myChartType = TTimeSeriesChartType.ctLDP;
                break;
                default:
                    myChartType = TTimeSeriesChartType.ctUndefined;
                break;
            }
            return myChartType;
        }
/******************************************************************************
* MODULE NAME ...  ReadFileData
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:
* Reads the file until the label record is found
*
* INPUT PARAMETERS:
*                    filePath: string                the stream to load from
*
* OUTPUT PARAMETERS:
*                    fileData: TStringList containing data content of file
*                    recordData: TStringList containing parsed header record (channel names)
*                    fileType:  TTimeSeriesFileType  ftCSV or ftReport
* RETURN VALUE:
*                    found : header record found true/false
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Initial implementation
* 2008.12.04  Kent Myers        Add test for file type
*******************************************************************************/
        public Boolean ReadFileData(string filePath, List<string> fileData, List<string> recordData,
                            ref TTimeSeriesFileType fileType)
        {
            const int RECORD_DATA = 6;

            int recordCount, iRecord;
            Boolean chartFound, deviceFound, eventFound, channelFound;
            char delim = ' ';
            string deviceField, eventField, channelField;
            GlobalString gs = new GlobalString();

            chartFound = false;       // assume valid header record not found

            // Each of these flags corresponds to a table in a CSV file

            deviceFound = false;
            eventFound = false;
            channelFound = false;

            if(File.Exists(filePath)) 
            {
                try
                {
                    gs.LoadFromFile(filePath,fileData);

                    fileType = GetFileType(fileData,ref delim);

                    iRecord = 0;  // index of current record in file
                    recordCount = fileData.Count;

                    if(recordCount <= 2)     // cannot have valid data
                    {
                        FChartType = TTimeSeriesChartType.ctUndefined;
                        return false;
                    }

                    // keywords used to identify each table in file

                    deviceField = "DeviceID";  // column name of device information table
                    channelField = "ChannelName";  // column name of channel table
                    eventField = "EventDate";  // column name of event table

                    if(fileType == TTimeSeriesFileType.ftCSV)    // get standard X and Y data columns
                    {
                        ReadFileCSVHeaderData(fileData, recordData);
                        iRecord = RECORD_DATA;   // record where non-header CSV tables start
                    }

                    // Read tables that are not chart data

                    while (iRecord < recordCount  )
                    {
                        gs.ParseString(delim,fileData[iRecord],recordData);
                        if(recordData.Count > 0) 
                        {
                            if(!deviceFound)   // no device information found yet
                            {
                                if(fileType == TTimeSeriesFileType.ftCSV) 
                                {
                                    if(gs.CompareText(recordData[0],deviceField))
                                    {
                                        ReadFileCSVDeviceData(ref iRecord,fileData,recordData);
                                        deviceFound = true;
                                    }
                                } 
                                else if(fileType == TTimeSeriesFileType.ftReport) 
                                {
                                    ReadFileReportDeviceData(ref iRecord,fileData,recordData,delim);
                                    deviceFound = true;
                                }
                            }
                            else if(!channelFound && 
                                   (fileType == TTimeSeriesFileType.ftCSV) && 
                                   (ChartType == TTimeSeriesChartType.ctLDP) &&
                                   gs.CompareText(recordData[0],channelField))
                            {
                                ReadFileCSVChannelData(ref iRecord,fileData,recordData,eventField);
                                channelFound = true;
                            }
                            else if(!eventFound && 
                                   (fileType == TTimeSeriesFileType.ftCSV) && 
                                   (ChartType == TTimeSeriesChartType.ctLDP) &&
                                   gs.CompareText(recordData[0],eventField))
                            {
                                ReadFileCSVEventData(ref iRecord,fileData,recordData);

                                eventFound = true;
                                iRecord--;  // set pointer back to re-read this record
                            }
                            else if(!chartFound &&
                                   gs.CompareText(recordData[0],ChartField))
                            {
                                chartFound = true;  // found start of chart data so stop reading
                                break;
                            }
                        }
                        iRecord++;
                    }
                }
                catch
                {
                    chartFound = false;
                    FChartType = TTimeSeriesChartType.ctUndefined;
                }
            }
            return chartFound;
        }
/************************************************************************
* NAME:       ReportFileHasStatusColumn
* FUNCTIONAL DESCRIPTION:   determine whether file has a status column
* INPUT PARAMETERS: fileData : TStringList   contains content of file
* OUTPUT PARAMETERS:
*                   count : integer   number of columns of data
* RETURN: result : boolean   true if file has Status column
************************************************************************/
        public Boolean ReportFileHasStatusColumn(List<string> fileData, out int count)
        {
            int iRecord, recordCount;
            List<string> recordData;
            Boolean hasStatus;
            GlobalString gs = new GlobalString();

            iRecord = 0;  // index of current record in file
            recordCount = fileData.Count;
            hasStatus = false;
            count = 0;

            recordData = new List<string>();
            try
            {
                //recordData.Delimiter = ' ';
                while (iRecord < recordCount)  
                {
                    gs.ParseString(' ',fileData[iRecord],recordData);
                    //recordData.DelimitedText := fileData[iRecord];
                    if(recordData.Count > 0) 
                    {
                        if(recordData[0] == ChartField) 
                        {
                            count = recordData.Count;
                            hasStatus = gs.CompareText(recordData[count-1],"STATUS");
//                            hasStatus = (UpperCase(recordData[count-1]) == "STATUS");
                            break;
                        }
                    }
                    iRecord++;
                }

                //finally
                //  recordData.Free;
            }
            catch 
            {
            }
            return hasStatus;
        }
/************************************************************************
* NAME:       ReportType
* FUNCTIONAL DESCRIPTION:   return type of report file
* INPUT PARAMETERS: fileData : TStringList   contains content of file
* OUTPUT PARAMETERS:
* RETURN: result : result : TTimeSeriesChartType  (ctSSI, ctLDP, ctUnknown)
************************************************************************/
        public TTimeSeriesChartType ReportType(string filePath)
        {
            List<string> recordData;
            List<string> fileData;
            TTimeSeriesChartType reportType;
            GlobalString gs = new GlobalString();

            reportType = TTimeSeriesChartType.ctUndefined;

            if(File.Exists(filePath)) 
            {
                fileData = new List<string>();
                recordData = new List<string>();
                try
                {
                    gs.LoadFromFile(filePath,fileData);
                    //fileData.LoadFromFile(filePath);
                    reportType = GetReportType(fileData,recordData);
                    //finally
                    //  recordData.Free;
                    //  fileData.Free;
                }
                catch
                {
                }
            }
            return reportType;
        }
/************************************************************************
* NAME:       GetReportType
* FUNCTIONAL DESCRIPTION:   determine whether file has a SSI report content
* INPUT PARAMETERS: fileData : TStringList   contains content of file
*                   recordData : TstringList  use for parsing file
* OUTPUT PARAMETERS:
* RETURN: result :  fileType : TTimeSeriesFileType
************************************************************************/
        public TTimeSeriesChartType GetReportType(List<string> fileData, List<string> recordData )
        {
            int iRecord;
            TTimeSeriesChartType reportType;
            char delim = ' ';
            GlobalString gs = new GlobalString();

            iRecord = 0;  // index of current record in file
            reportType  = TTimeSeriesChartType.ctUndefined;  // assume unknown type

            GetFileType(fileData,ref delim);
//            recordData.Delimiter := delim;

            while (iRecord < fileData.Count  )
            {
                gs.ParseString(delim,fileData[iRecord],recordData);
                //recordData.DelimitedText := fileData[iRecord];
                if(recordData.Count > 0) 
                {
                    if(recordData[0] == ChartField)   // reached label header (can't be SSI)
                    {
                        reportType = TTimeSeriesChartType.ctLDP;  // must be LDP
                        break;
                    }
                    else if(gs.ContainsText(fileData[iRecord],"Voltage(%")) 
                    {
                        reportType = TTimeSeriesChartType.ctSSI;  // SSI type
                        break;
                    }
                    else if(iRecord > 20) 
                    {
                        break;  // no need to keep looking
                    }
                }
                iRecord++;
            }
            return reportType;
        }
/************************************************************************
* NAME:       ReadFileCSVDeviceData
* FUNCTIONAL DESCRIPTION:    Read data describing the device (MID, FID, TID, settings)
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public void ReadFileCSVHeaderData(List<string> fileData, List<string> recordData)
        {
            const int RECORD_VERSION = 1;
            const int RECORD_STD_COL = 3;
            const int RECORD_DATA_COL = 5;
            const int FIELD_CHARTNAME = 0;
            //const int FIELD_CHARTVERSION = 1;
            const int FIELD_CHARTDATE = 2;
            const int FIELD_CHARTTIME = 3;
            const int FIELD_CHARTFIELD = 4;
            const int FIELD_TIMESOURCE = 5;
            const int FIELD_CHARTCOUNT = 5;  // TIMESOURCE was added so it isn't counted
            const string HEADER_RECORD_MSG = "Insufficient header records in CSV file.";
            const string HEADER_DATA_MSG = "Insufficient header data in CSV file.";
            const string HEADER_CHEADER_MSG = "Invalid header column.";
            const string HEADER_CDATA_MSG = "Invalid data column.";
            const string HEADER_CHART_TYPE_MSG = "Unknown chart type in data file";

            int i, iValue = 0;
            string myDate, myTime;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings();

            InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormat);
            myFormat.ShortDateFormat = "MM/dd/yy";   // must be standard English

            // are there enough records?
            if(RECORD_VERSION > (fileData.Count - 1)) 
            {
                throw new Exception(HEADER_RECORD_MSG);
            }
            gs.ParseString(',',fileData[RECORD_VERSION],recordData);
//            recordData.CommaText := fileData[RECORD_VERSION];    // get Chart record name

            // are there enough fields in the record?
            if(recordData.Count < FIELD_CHARTCOUNT)    // original field count
            {
                throw new Exception(HEADER_DATA_MSG);
            }

            FChartName = recordData[FIELD_CHARTNAME];

            myDate = recordData[FIELD_CHARTDATE];
            myTime = recordData[FIELD_CHARTTIME];

            FChartDateTime = gs.StrToDateTime(myDate + ' ' + myTime, myFormat);

            FChartField = recordData[FIELD_CHARTFIELD];

            //Timesource is a new field, so test to see if it is available

            if(recordData.Count > FIELD_CHARTCOUNT)    // added timesource to header
            {
                FTimeSource = recordData[FIELD_TIMESOURCE];
            }
            else
            {
                FTimeSource = "Unknown";      // In case timesource is not present
            }

            FChartType = GetChartType(FChartName);

            if(ChartType == TTimeSeriesChartType.ctUndefined)    // is there a valid chart type in file?
            {
                throw new Exception(HEADER_CHART_TYPE_MSG);
            }

            gs.ParseString(',',fileData[RECORD_STD_COL],recordData);
//            recordData.DelimitedText := fileData[RECORD_STD_COL];  // Indep. var. record

            // get column indices for chart table that are indep. vars (rec, Date, Time, etc.)

            SetLengthHeaderColumns(recordData.Count);
            for (i = 0; i<recordData.Count; i++)
            {
                if(!gs.TryStrToInt(recordData[i],ref iValue)) 
                {
                    throw new Exception(HEADER_CHEADER_MSG);
                }
                else
                {
                    FHeaderColumns[i] = iValue;
                }
            }

            // get column indices for chart table that are dependent vars (channels)

            gs.ParseString(',',fileData[RECORD_DATA_COL],recordData);
//            recordData.DelimitedText := fileData[RECORD_DATA_COL]; // Dependent var. data columns

            SetLengthDataColumns(recordData.Count);
            for (i = 0; i<recordData.Count; i++) 
            {
                if(!gs.TryStrToInt(recordData[i],ref iValue)) 
                {
                    throw new Exception(HEADER_CDATA_MSG);
                }
                else
                {
                    FDataColumns[i] = iValue;
                }
            }
        }
/************************************************************************
* NAME:       ReadFileCSVDeviceData
* FUNCTIONAL DESCRIPTION:    Read data describing the device (MID, FID, TID, settings)
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public void ReadFileCSVDeviceData(ref int iRecord, List<string> fileData, List<string> recordData)
        {
            const int DEVICE_ID = 0;
            const int DEVICE_FID = 1;
            const int DEVICE_TID = 2;
            const int DEVICE_ID_COUNT = 3;

            int i;
            List<string> settingList;
            TTimeSeriesSettingRecord settingRecord;
            GlobalString gs = new GlobalString();

            SettingTable.Clear();
            settingList = new List<string>();
            try
            {
                // read setting names from incoming recordData
                for (i = DEVICE_ID_COUNT; i<recordData.Count; i++)     // skip ID's
                {
                    settingList.Add(recordData[i]);
                }
                iRecord++;

                //    recordData.CommaText := fileData[iRecord];  // next record
                gs.ParseString(',',fileData[iRecord],recordData);
                if(recordData.Count < DEVICE_ID_COUNT) 
                {
                    throw new Exception("Insufficient device data in CSV file.");
                }

                FMeterID = recordData[DEVICE_ID];     // get device ID's
                FMeterFID = recordData[DEVICE_FID];
                FMeterTID = recordData[DEVICE_TID];

                // read setting values
                for (i = DEVICE_ID_COUNT; i< recordData.Count; i++)   // skip ID's
                {
                    settingRecord = new TTimeSeriesSettingRecord();
                    settingRecord.SettingName = settingList[i-DEVICE_ID_COUNT];
                    settingRecord.SettingValue = gs.NonLocaleStrToFloat(recordData[i]);
                    FSettingTable.Add(settingRecord);
                }
            }
            catch
            {
            }
            //finally
            //  settingList.Free;
            //end;
        }
/************************************************************************
* NAME:       ReadFileReportDeviceData
* FUNCTIONAL DESCRIPTION:    Read data describing the device (MID, FID, TID, settings)
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
 *                   delim : char delimiter
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public void ReadFileReportDeviceData(ref int iRecord, List<string> fileData, List<string> recordData, char delim)
        {
            const string DATA_COLUMN_MSG = "Insufficient data columns in file.";
            const string DATA_RECORD_MSG = "Insufficient data records in file.";
            const string INVALID_FILE_MSG = "Invalid file.";
            const string CHART_TYPE_MSG = "Unknown chart type.";
            const string DATE_STRING = "Date:";
            const string TIME_STRING = "Time:";
            const string TIMESOURCE_STRING = "Time Source:";
            const string VBASE_STRING = "VBASE";
            const int VBASE_OFFSET = 8;

            int pos1,pos2;
            TTimeSeriesSettingRecord settingRecord;
            int i;
            string myDate, myTime;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings();

            iRecord--;   // start one record back
            pos1 = 0;
            while(pos1 <= 0)   // scan for record containing date
            {
                iRecord++;
                if(iRecord >= fileData.Count) 
                {
                    throw new Exception(INVALID_FILE_MSG);
                    //return;
                }
                pos1 = fileData[iRecord].IndexOf(DATE_STRING); // find date
                //    pos1 = AnsiPos(DATE_STRING,fileData[iRecord]); // find date
            }
            FMeterID = fileData[iRecord].Substring(0,pos1-1).Trim();
            //  FMeterID = Trim(AnsiLeftStr(fileData[iRecord],pos1-1));
            pos2 = fileData[iRecord].IndexOf(TIME_STRING); // find time
            //  pos2 = AnsiPos(TIME_STRING,fileData[iRecord]);  // find time

            if(pos2 >= 0)  // found time
            {
                myDate = fileData[iRecord].Substring(pos1+DATE_STRING.Length,pos2-pos1-DATE_STRING.Length).Trim();
                myTime = fileData[iRecord].Substring(pos2 + TIME_STRING.Length).Trim();
//                myTime = fileData[iRecord].Substring(fileData[iRecord].Length - pos2 - TIME_STRING.Length).Trim();
                //myDate := Trim(AnsiMidStr(fileData[iRecord],
                //                  pos1+Length(DATE_STRING),
                //                  pos2-pos1-Length(DATE_STRING)));
                //myTime := Trim(AnsiRightStr(fileData[iRecord],
                //                  Length(fileData[iRecord])-pos2-Length(TIME_STRING)));
            }
            else          // Missing time
            {
                myDate = fileData[iRecord].Substring(pos1 + DATE_STRING.Length).Trim();
 //               myDate = fileData[iRecord].Substring(fileData[iRecord].Length - pos1 - DATE_STRING.Length, fileData[iRecord].Length);
                //myDate = Trim(AnsiRightStr(fileData[iRecord],Length(fileData[iRecord])
                //              - pos1 - Length(DATE_STRING)));
                myTime = "";
            }

            switch(FChartType) {
                case TTimeSeriesChartType.ctSSI:     // Read English date/time
                    InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormat);
                    FChartDateTime = gs.StrToDateTime(myDate + ' ' + myTime, myFormat);
                break;
                case TTimeSeriesChartType.ctLDP:     // Read native date/time
                    //if(AnsiContainsStr(myDate,'/')) then  // English date convention
                    if(gs.ContainsText(myDate,"/")) 
                    {
                        InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormat);
                    }
                    else                             // International date convention
                    {
                        InitDateTimeFormat(TDateFormatType.ftInternational, ref myFormat);
                    }
                    myDate = myDate + " " + myTime;
                    FChartDateTime = gs.StrToDateTime(myDate, myFormat);
                break;
                default:
                    System.Diagnostics.Debug.Assert(false,"Invalid chart type");
                break;
            }

            iRecord++;

            //pos1 := AnsiPos(TIMESOURCE_STRING,fileData[iRecord]);  // find Time source
            pos1 = fileData[iRecord].IndexOf(TIMESOURCE_STRING);  // find Time source
            if(pos1 >= 0)    // Found a time source
            {
                //    FTimeSource = Trim(AnsiMidStr(fileData[iRecord],pos1 + 13,8));
                FTimeSource = fileData[iRecord].Substring(pos1+13,1).Trim();
                if(FTimeSource[0].ToString().ToUpper() == "I") 
                {
                    FTimeSource = "Internal";
                }
                else
                {
                    FTimeSource = "External";
                }

                //FMeterTID := Trim(AnsiLeftStr(fileData[iRecord],pos1-1));
                FMeterTID = fileData[iRecord].Substring(0,pos1-1).Trim();
            }
            else            // No time source, just use entire record as TID
            {
                FTimeSource = "Unknown";
                FMeterTID = fileData[iRecord].Substring(0,Math.Min(fileData[iRecord].Length,40)).Trim();
                //FMeterTID = Trim(AnsiLeftStr(fileData[iRecord],40));
            }

            iRecord+=2;
            if(iRecord > fileData.Count) 
            {
                throw new Exception(DATA_RECORD_MSG);
            }

            switch( FChartType )
            {
                case TTimeSeriesChartType.ctSSI:

                    FChartName = "VSSI";
                    pos1 = fileData[iRecord].IndexOf(' ');
                    if(pos1 >= 0)     // there is a separator
                    {
                        //FMeterFID := Trim(AnsiMidStr(fileData[iRecord],5,pos1));
                        FMeterFID = fileData[iRecord].Substring(5,pos1).Trim();
                        }
                    else            // just use entire record
                    {
                        FMeterFID = fileData[iRecord].Trim();
                    }
                    if(FMeterFID.Length == 0) 
                    {
                        FMeterFID = "Unknown";
                    }
                    // Look ahead to get VBASE
                    for (i = 0; i<VBASE_OFFSET-1; i++) 
                    {
                        gs.ParseString(delim,fileData[iRecord+i],recordData);
                        if((recordData.Count > 0) && (recordData[0].ToString() == "#")) 
                        {
                            gs.ParseString(delim,fileData[iRecord+i+2],recordData);// skip blank record
                            break;
                        }
                    }
                    settingRecord = new TTimeSeriesSettingRecord();
                    FSettingTable.Add(settingRecord);
                    settingRecord.SettingName = VBASE_STRING;
                    if(recordData.Count < 3) 
                    {
                    throw new Exception(DATA_COLUMN_MSG);
                    }
                    settingRecord.SettingValue = gs.NonLocaleStrToFloat(recordData[recordData.Count-3]);
                break;

                case TTimeSeriesChartType.ctLDP:

                    FChartName = "LDP";
                    pos1 = fileData[iRecord].IndexOf("Function:");
                    if(pos1 >= 0) 
                    {
                        //FMeterFID := Trim(AnsiLeftStr(filedata[iRecord],pos1-1));
                        FMeterFID = fileData[iRecord].Substring(0,pos1-1).Trim();
                    }
                    else
                    {
                        if(fileData[iRecord].IndexOf(ChartField) >= 0)   // missing FID record
                        {
                            FMeterFID = "Unknown";
                            iRecord--;  // record contains data column labels
                        }
                        else
                        {
                            pos1 = fileData[iRecord].IndexOf(' ');
                            if (pos1 >= 0)
                            {
                                FMeterID = fileData[iRecord].Substring(0, pos1).Trim();
                            }
                            else
                            {
                                FMeterFID = fileData[iRecord].Trim();
                            }

                        }
                    }

                break;
                default:
                    throw new Exception(CHART_TYPE_MSG);
                //break;
            }
        }
/************************************************************************
* NAME:       ReadFileCSVChannelData
* FUNCTIONAL DESCRIPTION:    Read data describing the channels
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
*                   eventField : string   name that indicates start of event data
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public void ReadFileCSVChannelData(ref int iRecord, List<string> fileData, List<string> recordData, string eventField)
        {
            const int CHANNEL_FIELDS = 5;
            const string CHANNEL_MSG = "Insufficient fields in channel table record.";
            GlobalString gs = new GlobalString();

            TTimeSeriesLDPChannelRecord channelRecord;

            iRecord++;

            gs.ParseString(',',fileData[iRecord],recordData);
            LDPChannelTable.Clear();

            // read records until the Event Table or the Chart Table is found

            while ((recordData[0] != eventField) && (recordData[0] != ChartField)) 
            {
                if(recordData.Count < CHANNEL_FIELDS)    // test for valid channel record
                {
                    throw new Exception(CHANNEL_MSG);
                }
                channelRecord = new TTimeSeriesLDPChannelRecord();
                try
                {
                    channelRecord.ChannelName = recordData[0];
                    channelRecord.ChannelFunction = recordData[1];
                    channelRecord.ChannelUnits = recordData[2];
                    channelRecord.ChannelRules = recordData[3];
                    channelRecord.ChannelScale = gs.NonLocaleStrToFloat(recordData[4]);
                    iRecord++;
                    gs.ParseString(',',fileData[iRecord],recordData);  // get next record
                    LDPChannelTable.Add(channelRecord);
                }
                catch
                {
                }
            }
            // Decrement record index so that this record will be tested again
            // since the next function does not know which table was found
            iRecord--;
        }
/************************************************************************
* NAME:       ReadFileCSVEventData
* FUNCTIONAL DESCRIPTION:    Read data describing events
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************/
        public void ReadFileCSVEventData(ref int iRecord, List<string> fileData, List<string> recordData)
        {
            const int EVENT_FIELDS = 4;
            const string EVENT_MSG = "Insufficient fields in event table record.";

            TTimeSeriesLDPEventRecord eventRecord;
            string myDateTime;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings();

            InitDateTimeFormat(TDateFormatType. ftEnglish_US, ref myFormat);
            myFormat.ShortDateFormat = "MM/dd/yy";   // must be standard English

            iRecord++;
            gs.ParseString(',',fileData[iRecord],recordData);
            LDPEventTable.Clear();
            while (recordData[0] != ChartField) 
            {
                if(recordData.Count < EVENT_FIELDS)    // test for valid event record
                {
                    throw new Exception(EVENT_MSG);
                }
                eventRecord = new TTimeSeriesLDPEventRecord();
                try
                {
                    eventRecord.EventName = String.Format("#{0:d4}",LDPEventTable.Count+1);
                    myDateTime = recordData[0] + " " + recordData[1];
                    eventRecord.DateTime = gs.StrToDateTime(myDateTime, myFormat);
                    eventRecord.EventCode = recordData[2];
                    eventRecord.Description = recordData[3];
                    iRecord++;
                    gs.ParseString(',',fileData[iRecord],recordData); // get next record
                    LDPEventTable.Add(eventRecord);
                }
                catch
                {
                }
            }
        }
/************************************************************************
* NAME:       InitDateTimeFormat
* FUNCTIONAL DESCRIPTION:    Initialize date/time format record
* INPUT PARAMETERS: formatType : TDateFormatType  (ftEnglish_US, ftInternational)
*                   myFormat : TFormatSettings
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void InitDateTimeFormat(TDateFormatType formatType, ref TFormatSettings myFormat)
        {
            GlobalString gs = new GlobalString();
            gs.InitTFormatSettings(formatType, FDateFormat, ref myFormat);
        }
/************************************************************************
* NAME:       LoadData
* FUNCTIONAL DESCRIPTION:  loads the data portion of the record from the
*                          given stream using the given record size
* INPUT PARAMETERS:
*                 fileData: TStringList           list of data records
*                 xColumns: array of integer      list of independent var data colums
*                                                 (record, date, time, phase, status)
*                 yColumns:  array of integer      list of dependent data columns
* OUTPUT PARAMETERS: none
* RETURN:
*                 result: boolean   true if data ok, false if data bad
************************************************************************/
        public Boolean LoadData(List<string> fileData, int[] xColumns, int[] yColumns)
        {
            const string DATA_OVERWRITE = "Command Aborted, Data overwrite occurred.";
            const string DATA_INTERRUPT = "No Voltage Sag/Swell/Interruption Data";
            const string DATA_INVALID = "Invalid Data";
            const string DATA_FILE_INVALID = "Invalid file content.";
            const string DATA_TRIGGER = "*";
            const int CURRENT_START = 0;
            const int CURRENT_END = 4;
            const int VOLT_START = 5;
            const int VOLT_END = 7;
            const int MINIMUM_COLUMNS_SSI = 8;
            const int MINIMUM_COLUMNS_LDP = 4;

            GlobalString gs = new GlobalString();
            List<string> recordData;
            int i,j, iRecord;
            Boolean chartFound,status;
            int columnCount;
            TTimeSeriesFileType fileType;
            string myStatus;
            TFormatSettings myFormat = gs.InitTFormatSettings();
            char delim = ' ';
            TTimeSeriesLDPChannelRecord channelRecord;

            System.Diagnostics.Debug.Assert(fileData != null, DATA_FILE_INVALID);
            chartFound = false;

            recordData = new List<string>();
            try
            {
                fileType = GetFileType(fileData,ref delim);
                //recordData.Delimiter := delim;

                columnCount = 0;
                FDeviceType = TTimeSeriesDeviceType.dtSEL7XX;   // in case it is a relay SSI report

                if(fileType == TTimeSeriesFileType.ftCSV) 
                {
                    InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormat);
                    myFormat.ShortDateFormat = "MM/dd/yy";   // must be standard English
                }
                else
                {
                    if(ChartType == TTimeSeriesChartType.ctSSI) 
                    {
                        InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormat);
                    }
                    else
                    {
                        InitDateTimeFormat(TDateFormatType.ftInternational, ref myFormat);
                    }
                }

                // extract data from each data record
                iRecord = 0;
                while (iRecord < fileData.Count)
                {
                    // test for error messages in record
                    if(gs.ContainsText(fileData[iRecord],DATA_OVERWRITE)) 
                    {
                        iRecord++;
                        continue; // skip record
                    }
                    if(gs.ContainsText(fileData[iRecord],DATA_INVALID)) 
                    {
                        iRecord++;
                        continue; // skip record
                    }
                    if(gs.ContainsText(fileData[iRecord],DATA_INTERRUPT)) 
                    {
                        return false;
                    }

                    if (gs.ContainsText(fileData[iRecord], DATA_TRIGGER)) 
                    {
                        myStatus = "*";
                        fileData[iRecord] = fileData[iRecord].Replace("*"," ");  // replace "trigger" flag
                    }
                    else
                    {
                        myStatus = " ";
                    }

                    gs.ParseString(delim,fileData[iRecord],recordData);
                    //if(fileType = ftReport) then
                    //begin
                    //  recordData.DelimitedText := fileData[iRecord];  // Report
                    //end else
                    //begin
                    //  recordData.CommaText := fileData[iRecord];   // CSV
                    //end;

                    if(recordData.Count > 0)    // something is in the record
                    {
                        if(!chartFound && gs.CompareText(recordData[0],ChartField))
                        {
                            if(((ChartType == TTimeSeriesChartType.ctSSI) && (recordData.Count < MINIMUM_COLUMNS_SSI)) ||
                            ((ChartType == TTimeSeriesChartType.ctLDP) && (recordData.Count < MINIMUM_COLUMNS_LDP)))  // invalid column count
                            {
                                return false;
                            }
                            // 734 Meter and 3xx/4xx relays have different number of columns in SSI
                            if((ChartType == TTimeSeriesChartType.ctSSI) &&
                            ((recordData.Count-COL_TIME-1) < yColumns[yColumns.Length-1]))  // SEL4XX
                            {
                                for( i = CURRENT_START; i<=CURRENT_END; i++)  
                                {
                                    yColumns[i] = -1;  // deactivate current columns
                                }
                                for( i = VOLT_START; i<=VOLT_END; i++)  
                                {
                                    yColumns[i] -= VOLT_START;  // decrement voltage column indices
                                }
                            }

                            for(j = 0 ; j<yColumns.Length; j++) 
                            {
                                if(yColumns[j] < 0) continue;  // inactive column, skip it
                                RawDataLabels.Add(recordData[yColumns[j]]);
                            }

                            if((ChartType == TTimeSeriesChartType.ctLDP) && (fileType == TTimeSeriesFileType.ftReport))   // init channel table
                            {
                                LDPChannelTable.Clear();
                                for (j = 0; j<RawDataLabels.Count; j++) 
                                {
                                    channelRecord = new TTimeSeriesLDPChannelRecord();
                                    try
                                    {
                                        channelRecord.ChannelName = RawDataLabels[j];
                                        channelRecord.ChannelFunction ="Unknown";
                                        channelRecord.ChannelUnits = "Unknown";
                                        channelRecord.ChannelRules = String.Format("{0:f2}",1.0);  // regional language
                                        channelRecord.ChannelScale = 1.0F;
                                        LDPChannelTable.Add(channelRecord);
                                    }
                                    catch
                                    {
                                    }
                                }
                            }
                            chartFound = true;      // found header, remaining records are data
                        } 
                        else
                        {
                            if(chartFound)       // we are reading data records from now on
                            {
                                status = ReadChartRecord(fileType,recordData,myFormat,xColumns,yColumns,ref columnCount,ref myStatus);
                                if(!status) 
                                {
                                    return status;
                                }
                            }
                        }
                    }
                    else if(chartFound && (RawDataTable.Count > 0))   // encountered blank record in data, stop now
                    {
                        return true;
                    }
                    iRecord++;
                }
            }
            catch 
            {
            }
            //finally
            //  FreeAndNil(recordData);
            //end;
            return chartFound;
        }
/************************************************************************
* NAME:       ReadChartRecord
* FUNCTIONAL DESCRIPTION:    Read and parse a chart record and add to collection
* INPUT PARAMETERS:  fileType : TTimeSeriesFileType   type of file
*                    recordData : TStringList  content of present record
*                    myFormat : TFormatSettings  used to parse date/time strings
*                    standard : array of integer  column position of standard values
*                    columns : array of integer column position of data columns
* OUTPUT PARAMETERS:
*                    columnCount : expected number of columns
*                    deviceType : TTimeSeriesDeviceType (dtSEL3xx,...)
*                    myStatus : string  status field
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************/
        public Boolean ReadChartRecord(TTimeSeriesFileType fileType,
               List<string> recordData,
               TFormatSettings myFormat,
               int[] xColumns, int[] yColumns,
               ref int columnCount,
               ref string myStatus)
        {
            const string FIELD_COUNT_MSG = "Insufficient fields in record.";
            const string INVALID_RECORD_MSG = "Invalid record number.";
            const string INVALID_DATE_MSG = "Invalid date.";
            const string INVALID_PHASE_MSG = "Invalid phase.";
            const string INVALID_STATUS_MSG = "Invalid status.";
            const string INVALID_DATA_MSG = "Invalid channel data.";

            TTimeSeriesRecord dataRecord;
            Boolean valid;

            if(recordData.Count < 1)     // make sure something in record
            {
                throw new Exception(FIELD_COUNT_MSG);
            }

            if(recordData[0].ToString() == "#")    // found comment in data, stop scanning
            {
                return false;
            }
            else
            {
                if(columnCount == 0)      // not set yet
                {
                    columnCount = recordData.Count;  // remember column count
                    FDeviceType = DetermineDeviceType(fileType,columnCount,xColumns[xColumns.Length-1]);
                }
                if((columnCount != recordData.Count) ||       // inconsistent column count
                (recordData.Count < yColumns[yColumns.Length-1]))   // not enough data columns
                {
                    return false;
                }
                dataRecord = new TTimeSeriesRecord();

                try
                {
                    // Get Record number
                    valid = ReadChartRecordNumber(recordData, xColumns, ref dataRecord);
                    if(!valid) 
                    {
                        throw new Exception(INVALID_RECORD_MSG);
                    }

                    // Get Date
                    valid = ReadChartRecordDate(recordData, xColumns, fileType, myFormat, ref dataRecord);
                    if(!valid) 
                    {
                        throw new Exception(INVALID_DATE_MSG);
                    }

                    // Get Phase
                    valid = ReadChartRecordPhase(recordData, xColumns, ref dataRecord);
                    if(!valid) 
                    {
                        throw new Exception(INVALID_PHASE_MSG);
                    }

                    // Get Status
                    valid = ReadChartRecordStatus(recordData, xColumns, ref dataRecord);
                    if(!valid) 
                    {
                        throw new Exception(INVALID_STATUS_MSG);
                    }
                    myStatus = dataRecord.Status;     // return the status that was found

                    // Get Data
                    valid = ReadChartRecordData(recordData, yColumns, fileType, ref dataRecord);
                    if(!valid && (myStatus != "X"))   // X is invalid record marker
                    {
                        throw new Exception(INVALID_DATA_MSG);
                    }

                    RawDataTable.Add(dataRecord);
                }
                catch // dataRecord was already added to the RawDataTable
                {
                    return false;
                }
            }
            return true;
        }
/************************************************************************
* NAME:       ReadChartRecordNumber
* FUNCTIONAL DESCRIPTION:    Read the RECORD Number column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
*                    fileType : TTimeSeriesFileType   ftCSV, ftReport...
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************/
        public Boolean ReadChartRecordNumber(
               List<string> recordData,
               int[] xColumns,
               ref TTimeSeriesRecord dataRecord)
        {
            GlobalString gs = new GlobalString();

            if(xColumns[COL_RECORD] >= 0)     // extracting actual record number
            {
                if(!gs.TryStrToInt(recordData[xColumns[COL_RECORD]], ref dataRecord.RecordNumber)) 
                {
                return false;
                }
            }
            else                     // construct record number automatically
            {
                dataRecord.RecordNumber = RawDataTable.Count + 1;  // record not saved yet so add one to rec. no.
            }
            return true;
        }
/************************************************************************
* NAME:       ReadChartRecordDate
* FUNCTIONAL DESCRIPTION:    Read the DATE column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
*                    fileType : TTimeSeriesFileType   ftCSV, ftReport...
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************/
        public Boolean ReadChartRecordDate(
               List<string> recordData,
               int[] xColumns,
               TTimeSeriesFileType fileType,
               TFormatSettings myFormat,
               ref TTimeSeriesRecord dataRecord)
        {
            string myDateTime;
            Boolean valid;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormatEnglish = gs.InitTFormatSettings();

            if(xColumns[COL_DATE] >= 0)    // extract and record date/time
            {
                myDateTime = recordData[xColumns[COL_DATE]] + ' ' + recordData[xColumns[COL_TIME]];
                if(gs.ContainsText(myDateTime,"/"))   // English date
                {
                    InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormatEnglish);
                    if(fileType == TTimeSeriesFileType.ftCSV) 
                    {
                        myFormatEnglish.ShortDateFormat = "MM/dd/yy";   // must be standard English
                    }
                    dataRecord.DateTime = gs.StrToDateTime(myDateTime, myFormatEnglish);
                }
                else
                {                                     // International date
                    dataRecord.DateTime = gs.StrToDateTime(myDateTime, myFormat);
                }
                valid = IsValidDateTime(myDateTime);
                if(!valid)         // date time did not match any valid format
                {
                    return false;
                }
            }
            else
            {                       // use current date/time
            dataRecord.DateTime = DateTime.Now;
            }
            return true;
        }

/************************************************************************
* NAME:       ReadChartRecordPhase
* FUNCTIONAL DESCRIPTION:    Read the PHASE column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************/
        public Boolean ReadChartRecordPhase(
               List<string> recordData,
               int[] xColumns,
               ref TTimeSeriesRecord dataRecord)
        {
            const int PHASE_A = 1;
            const int PHASE_C = 3;

            int j,iColumn;

            if(xColumns[COL_PHASE] >= 0)     // extract Phase descriptor
            {
                iColumn = xColumns[COL_PHASE];    // usual column for PHASE
                switch( FDeviceType )
                {
                    case TTimeSeriesDeviceType.dtSEL3XX:

                    iColumn += RELAY_3XX_OFFSET;
                    break;
                    case TTimeSeriesDeviceType.dtSEL4XX:

                    iColumn += RELAY_4XX_OFFSET;
                    break;
                }

                dataRecord.Phase = recordData[iColumn];

                for( j = PHASE_A; j<PHASE_C-1 ; j++)
                {
                    if((dataRecord.Phase[j] != 'I') &&
                    (dataRecord.Phase[j] != 'O') &&
                    (dataRecord.Phase[j] != 'U') &&
                    (dataRecord.Phase[j] != '.')) 
                    {
                        return false;
                    }
                }
            }
            else
            {
                dataRecord.Phase = "";    // empty string
            }
            return true;
        }
/************************************************************************
* NAME:       ReadChartRecordStatus
* FUNCTIONAL DESCRIPTION:    Read the STATUS column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************/
        public Boolean ReadChartRecordStatus(
               List<string> recordData,
               int[] xColumns,
               ref TTimeSeriesRecord dataRecord)
        {
            int iColumn, index;
            string myStatus;
            const string STATUS = "RPFEMSDX"; // possible content of status column

            if(xColumns[COL_STATUS] >= 0)     // extract status descriptor
            {
                iColumn = xColumns[COL_STATUS];     // usual column for STATUS
                switch( FDeviceType)
                {
                    case TTimeSeriesDeviceType.dtSEL3XX:
                        iColumn += RELAY_3XX_OFFSET;
                    break;
                    case TTimeSeriesDeviceType.dtSEL4XX:
                        iColumn += RELAY_4XX_OFFSET;
                    break;
                }
                myStatus = recordData[iColumn];
                if(myStatus.Length > 0) 
                {
                    if(FChartType == TTimeSeriesChartType.ctSSI)    // extract SSI Phase descriptor
                    {
                        index = STATUS.IndexOf(myStatus);
                        //index = AnsiIndexStr(myStatus,['R','P','F','E','M','S','D','X']);
                        if(index < 0) 
                        {
                            return false;
                        }
                    }
                }
                else
                {
                    myStatus = "...";
                }
            }
            else
            {
                myStatus = "...";
            }
            dataRecord.Status = myStatus;
            return true;
        }
/************************************************************************
* NAME:       ReadChartRecordData
* FUNCTIONAL DESCRIPTION:    Read the DATA columns from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    yColumns : array of integer  column position of dep. data values
*                    fileTYpe : TTimeSeriesFileType (ftCSV, ftReport,...)
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************/
        public Boolean ReadChartRecordData(
               List<string> recordData,
               int[] yColumns,
               TTimeSeriesFileType fileType,
               ref TTimeSeriesRecord dataRecord)
        {
            int j,index, iColumn,count;
            Boolean status;
            string myValue;
            GlobalString gs = new GlobalString();

            status = true;  // assume read is successful
            // Extract the active dependent variable data columns
            count = 0;
            index = 0;
            for (j = 0; j<yColumns.Length; j++) 
            {
                if(yColumns[j] > 0) 
                {
                    count++;
                }
            }

            dataRecord.Value = new float[count+1];
            for (j = 0; j< yColumns.Length; j++) 
            {
                iColumn = yColumns[j];  // default column location
                if(iColumn < 0) continue;  // inactive column
                if((FChartType == TTimeSeriesChartType.ctSSI) && (FDeviceType == TTimeSeriesDeviceType.dtSEL3XX)) 
                {
                    if(j > RELAY_3XX_START) 
                    {
                        iColumn += RELAY_3XX_OFFSET;
                    }
                }

                myValue = gs.InternalizeData(recordData[iColumn]);
                if (!gs.TryNonLocaleStrToFloat(myValue,ref dataRecord.Value[index]))
                {
                    dataRecord.Value[index] = 0.0F;
                    status = false;
                }
                index++;
            }
            return status;
        }
/************************************************************************
* NAME:     DetermineDeviceType
* FUNCTIONAL DESCRIPTION:   Determine device type from column count
* INPUT PARAMETERS: fileType : TTimeSeriesFileType  ftReport, ftCSV
*                   columnCount : integer number of columns of data found
*                   lastColumn  : integer  last data column expected
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesDeviceType (dtSEL7XX, ...)
************************************************************************/
        public TTimeSeriesDeviceType DetermineDeviceType(TTimeSeriesFileType fileType,
                 int columnCount, int lastColumn)
        {
            if((fileType == TTimeSeriesFileType.ftReport) && (FChartType == TTimeSeriesChartType.ctSSI)) 
            {
                if((columnCount-1) > lastColumn)    // must be a "351 Relay" SSI report
                {
                    FDeviceType = TTimeSeriesDeviceType.dtSEL3XX;   // more columns in relay SSI report
                }
                else if((columnCount-1) < lastColumn)    // must be a "487 Relay" SSI report
                {
                    FDeviceType = TTimeSeriesDeviceType.dtSEL4XX;   // Fewer columns in relay SSI report
                }
                else
                {
                    FDeviceType = TTimeSeriesDeviceType.dtSEL7XX;   // standard number of columns (based on 734)
                }
            }
            else
            {
                FDeviceType = TTimeSeriesDeviceType.dtSEL7XX;
            }
            return FDeviceType;
        }
/************************************************************************
* NAME:     IsValidDateTime
* FUNCTIONAL DESCRIPTION:   test date/time string for validity
* INPUT PARAMETERS: myDateTime : string  date/time string read from file
* OUTPUT PARAMETERS: none
* RETURN: result : boolean  true if date/time is valid otherwise false
************************************************************************/
        public Boolean IsValidDateTime(string myDateTime)
        {
            return true;  // accept any  date/time format for now
        }
/******************************************************************************
* MODULE NAME ...  CompareRecords
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Comparison function used by sort method
*
* INPUT PARAMETERS:
*                 Item1,Item2: Pointer           ptr to data objects being compared
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result : integer -1 if Item1 < Item2
*                                     0 if Item1 = Item2
*                                     1 if Item1 > Item2
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Initial implementation
*******************************************************************************/
        //public int CompareRecords(void Item1, void Item2)
        //{
        //    return CompareDateTime(TTimeSeriesRecord(Item1).DateTime,
        //                        TTimeSeriesRecord(Item2).DateTime);
        //}
    }

}
