/*****************************************************************************
* UNIT DESCRIPTION:  Base class for charting functions
* SPEC REF: (if used)
* DESIGN REF: (if used)
* 
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************/
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using SELGlobalStringMethods;
using Steema.TeeChart;
//#if WPF
using Steema.TeeChart.WPF;
using Steema.TeeChart.WPF.Tools;
using Steema.TeeChart.WPF.Themes;
using System.Windows.Controls;
using Steema.TeeChart.WPF.Styles;
using Microsoft.Win32;
using Steema.TeeChart.WPF.Drawing;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Input;
using System.Windows;
//#endif

namespace SELChartClasses
{
    public struct TLimits
    {
        public int limit1;
        public int limit2;
    }

    public class TTimeSeriesChart
    {
        private int FMouseX,FMouseY; // last location of mouse after mouse down/move
        private int FChartVariableCount; // number of variables in chart (including CBEMA)
        private Boolean FChartHasEvents; // true if event limits are present
        private int FEventLastSortedColumn;
        private Boolean FEventAscending;
        private double FChartXDelta;  // interval between chart labels

        private TTimeSeriesDataTable FEventData;  // Table containing raw data

        private TChart FChartDisplay;  // active chart
        private CursorTool FCursorTool;  // used for data query with mouse
        private Boolean FChartRotate;  // rotation is active
        private Boolean FChartTranslate;  // translation is active
        private Boolean FChartZoom;  // zoom is active
        private Boolean FChartThickness;  // 3D depth thickness change is active
        private Boolean FChartActive;  // dynamic operation active (rotate,...)

        public Boolean ChartRotate
        {
            get{return FChartRotate;}
            set{FChartRotate = value;}
        }

        public Boolean ChartTranslate
        {
            get{return FChartTranslate;}
            set{FChartTranslate = value;}
        }

        public Boolean ChartZoom
        {
            get{return FChartZoom;}
            set{FChartZoom = value;}
        }

        public Boolean ChartThickness
        {
            get{return FChartThickness;}
            set{FChartThickness = value;}
        }

        public Boolean ChartActive
        {
            get{return FChartActive;}
            set{FChartActive = value;}
        }

        public double ChartXDelta
        {
            get{return FChartXDelta;}
            set{FChartXDelta = value;}
        }

        public TChart ChartDisplay
        {
            get{return FChartDisplay;}
            set{FChartDisplay = value;}
        }
  
        public CursorTool CursorTool
        {
            get{return FCursorTool;}
            set{FCursorTool = value;}
        }
 
        public int MouseX
        {
            get{return FMouseX;}
            set{FMouseX = value;}
        }
      
        public int MouseY
        {
            get{return FMouseY;}
            set{FMouseY = value;}
        }

        public int ChartVariableCount
        {
            get{return FChartVariableCount;}
            set{FChartVariableCount = value;}
        }

        public Boolean ChartHasEvents
        {
            get{return FChartHasEvents;}
            set{FChartHasEvents = value;}
        }

        public int EventLastSortedColumn
        {
            get{return FEventLastSortedColumn;}
            set{FEventLastSortedColumn = value;}
        }

        public Boolean EventAscending
        {
            get{return FEventAscending;}
            set{FEventAscending = value;}
        }

        public TTimeSeriesDataTable EventData
        {
            get{return FEventData;}
            set{FEventData = value;}
        }

/************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************/
        public TTimeSeriesChart()
        {
            EventData = new TTimeSeriesDataTable();
            FChartXDelta = 10.0;
            FChartHasEvents = false;
            FEventAscending = true;
            FEventLastSortedColumn = -1;
        }
/************************************************************************
* NAME:   Destroy
* FUNCTIONAL DESCRIPTION: class destructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************/
//destructor TTimeSeriesChart.Destroy();
//{
//  FreeAndNil(FEventData);
//  inherited Destroy();
//}
/************************************************************************
* NAME:   GetChartFilename
* FUNCTIONAL DESCRIPTION: Virtual base class
* INPUT PARAMETERS:  aChartType : THMIChartType (type of chart)
*                    aFileType : TTimeSeriesFileType  (type of file)
*                    ExportPath : default path to file
*                    aID        : ID of meter
* OUTPUT PARAMETERS: none
* RETURN:            filePath : string  path to file that was saved
************************************************************************/
        public virtual string GetChartFilename(TTimeSeriesFileType aFileType, string ExportPath, string aID)
        {
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
            return " ";
        }
/************************************************************************
* NAME:   TrimFileName
* FUNCTIONAL DESCRIPTION:  trim filename from the end of a file path name
* INPUT PARAMETERS: filename : string  is the original filename
* OUTPUT PARAMETERS: none
* RETURN:             result : string is the trimmed pathname
************************************************************************/
        public string TrimFileName(string filename)
        {
            string newFilename;
            int myPos;

            newFilename = filename;
            myPos = newFilename.LastIndexOf('/'); // look for last backslash
            if(myPos > 0)
            {
                newFilename.Remove(myPos+1,filename.Length); // leave a trailing backslash
            }
            return newFilename;
        }
/************************************************************************
* NAME:   SaveStringToFile
* FUNCTIONAL DESCRIPTION: Save string to given file name
* INPUT PARAMETERS:
*                    filePath : string  name of file to save string to
*                    aString : string   string to save
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************/
        public void SaveStringToFile(string filePath, string aString)
        {
            // Save data in file so that it may be read back as individual records

            if((filePath == null) || (filePath.Length == 0)) return;
            if((aString == null) || (aString.Length == 0)) return;

            StreamWriter fileWriter = null;
            try 
            {
                fileWriter = File.CreateText(filePath);
                fileWriter.Write(aString);
            }
            finally
            {
                if(fileWriter != null) fileWriter.Close();
            }
        }
/************************************************************************
* NAME:   RenameOldFile
* FUNCTIONAL DESCRIPTION: Test for old file and add a numeric extension
* INPUT PARAMETERS:
*                    filePath : string  name of file to rename
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************/
        public void RenameOldFile(string filePath)
        {
            const int EXTENSION_LENGTH = 3;

            //myFile : TextFile;
            string newFileName;
            int count, ipos;

            if(File.Exists(filePath))     // test for existing file with same name and rename it
            {
                count = 0;
                newFileName = filePath;
                while(File.Exists(newFileName))    // strip off counter extension
                {
                    count++;
                    ipos = newFileName.IndexOf('.');   // find start of first extension
                    newFileName = filePath.Substring(ipos+EXTENSION_LENGTH);  // restore original filename
                    newFileName = newFileName + String.Format("{0:d}",count);  // append counter
                }
                System.IO.File.Move(filePath,newFileName);
            }
        }
/************************************************************************
* NAME:  InitChart
* FUNCTIONAL DESCRIPTION: Initialize chart class members before importing report file
* INPUT PARAMETERS: fileType : TTimeSeriesFileType   type of file (ftReport,...)
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public virtual void InitChart(TTimeSeriesFileType fileType)
        {
            EventData.ChartType = TTimeSeriesChartType.ctInvalid;
            EventData.ChartName = "Unknown";
            EventData.ChartVersion = "1.0";

            // assign default separator identifying chartable data
            if(fileType == TTimeSeriesFileType.ftReport) 
            {
                EventData.ChartField = "#";
            }
            else
            {
                EventData.ChartField = "ChartField";
            }
        }
/************************************************************************
* NAME:    AppendCSVHeader
* FUNCTIONAL DESCRIPTION:  Append header records for CSV file to string
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: MeterBuffer : string  file content
* RETURN: result : boolean   status of append (true = success)
************************************************************************/
        public virtual Boolean AppendCSVHeader(ref string MeterBuffer)
        {
            const string CRLF = "\r\n";     // carriage return line feed

            int i;      // indices
            TTimeSeriesSettingRecord settingRecord;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings();
            string myDate,myTime;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");

            EventData.InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormat);
            myFormat.ShortDateFormat = "MM/dd/yy";

            //With FEventData do
            //{

            System.Diagnostics.Debug.Assert(EventData.RawDataLabels != null,"RawDataLabels is nil");
            System.Diagnostics.Debug.Assert(EventData.RawDataTable != null,"RawDataTable is nil");
            System.Diagnostics.Debug.Assert(EventData.IndexTable != null,"IndexTable is nil");

            try
            {
                myDate = gs.DateToStr(FEventData.ChartDateTime,myFormat);  // use setting date format
                myTime = String.Format("{0:hh:mm:ss.fff}",EventData.ChartDateTime);
                MeterBuffer = MeterBuffer +
                    "FileType, Version, Date, Time, ChartField, TimeSource" + CRLF +
                    EventData.ChartName + ", " + EventData.ChartVersion + ", " +
                    myDate + ", " + myTime + ", " +
                    EventData.ChartField + ", " + EventData.TimeSource + CRLF;

                // Export header columns (indep. vars)

                if(EventData.HeaderColumnArray.Length < 5) 
                {
                    throw new System.Exception("Insufficient number of independent variables.");
                }

                MeterBuffer = MeterBuffer + "RecNum, Date, Time, Phase, Status" + CRLF;
                for (i = 0; i < EventData.HeaderColumnArray.Length; i++) 
                {
                    if(i < (EventData.HeaderColumnArray.Length) - 1)
                    {
                        MeterBuffer = MeterBuffer + String.Format("{0:d}, ",EventData.HeaderColumns[i]);
                    }
                    else
                    {
                        MeterBuffer = MeterBuffer + String.Format("{0:d}",EventData.HeaderColumns[i]) + CRLF;
                    }
                }

                // Export data columns

                if(EventData.DataColumnArray.Length < 1) 
                {
                    throw new System.Exception("Insufficient number of dependent variables.");
                }

                // Channel names
                for (i = 0; i<EventData.RawDataLabels.Count; i++)
                {
                    if(i < (EventData.RawDataLabels.Count-1)) 
                    {
                        MeterBuffer = MeterBuffer + EventData.RawDataLabels[i] + ", ";
                    }
                    else
                    {
                        MeterBuffer = MeterBuffer + EventData.RawDataLabels[i] + CRLF;
                    }
                }

                // data column numbers (dependent vars)
                for( i = 0; i<EventData.DataColumnArray.Length; i++) 
                {
                    if(i < EventData.DataColumnArray.Length-1)
                    {
                        MeterBuffer = MeterBuffer + String.Format("{0:d, ",EventData.DataColumns[i]);
                    }
                    else
                    {
                        MeterBuffer = MeterBuffer + String.Format("{0:d}",EventData.DataColumns[i]) + CRLF;
                    }
                }

                // Table of Device information

                MeterBuffer = MeterBuffer + "DeviceID, FID, TID";
                for (i = 0; i< EventData.SettingTable.Count; i++)
                {
                    settingRecord = (EventData.SettingTable[i] as TTimeSeriesSettingRecord);
                    MeterBuffer = MeterBuffer + ", " + settingRecord.SettingName;
                }
                MeterBuffer = MeterBuffer + CRLF;

                MeterBuffer = MeterBuffer + String.Format("{\"{0:s}\", \"{1:s}\", \"{2:s}\"}",
                EventData.MeterID,
                EventData.MeterFID,
                EventData.MeterTID);
                for (i = 0; i< EventData.SettingTable.Count; i++)
                {
                    settingRecord = (EventData.SettingTable[i] as TTimeSeriesSettingRecord);
                    MeterBuffer = MeterBuffer + ", " + gs.NonLocaleFloatToStr(settingRecord.SettingValue);
                }
                MeterBuffer = MeterBuffer + CRLF;
            }
            catch
            {
                return false;
            }
            //}
            return true;
        }
/************************************************************************
* NAME:    AppendCSVData
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      MeterBuffer : string  file content
* RETURN:                 result : boolean   true if data appended, otherwise false
************************************************************************/
        public virtual Boolean AppendCSVData(ref string MeterBuffer)
        {
            // intentionally empty
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
            return false;
        }
/************************************************************************
* NAME:    AppendReport
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   MeterBuffer : string  file content
* RETURN:              result : boolean   true if data appended, otherwise false
************************************************************************/
        public virtual Boolean AppendReport(out string MeterBuffer)
        {
            // empty function
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
            MeterBuffer = " ";
            return false;
        }
/************************************************************************
* NAME:    InitChartVariableCount
* FUNCTIONAL DESCRIPTION:
* Set the chart variable count based on chart type and event count
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public virtual void InitChartVariableCount()
        {
            // update chart variable count to include CBEMA curves if appropriate

            System.Diagnostics.Debug.Assert(FEventData != null,"EventData is null");
            System.Diagnostics.Debug.Assert(FEventData.RawDataLabels != null,"RawDataLabels is null");

            FChartVariableCount = EventData.RawDataLabels.Count;  // number of channels to display
            FChartHasEvents = false;  // assume event limits not displayed
        }

/************************************************************************
* NAME:    InitChartActiveChannels
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public virtual void InitChartActiveChannels()
        {
            // intentionally empty
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
        }
/************************************************************************
* NAME:    InitChartActiveChannelsDefault
* FUNCTIONAL DESCRIPTION:   Virtual base class method
 INPUT PARAMETERS:    reset : boolean  
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public virtual void InitChartActiveChannelsDefault(Boolean reset)
        {
            // intentionally blank
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
        }
/************************************************************************
* NAME:    SaveChartActiveChannels
* FUNCTIONAL DESCRIPTION:  Virtual base class method
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public virtual void SaveChartActiveChannels()
        {
            // intentionally empty
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
        }
/************************************************************************
* NAME:    InitChartActiveEvents
* FUNCTIONAL DESCRIPTION:  Set the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public void InitChartActiveEvents()
        {
            int i;
            TTimeSeriesIndexRecord indexRecord;

            System.Diagnostics.Debug.Assert(FEventData != null,"EventData is nil");
            System.Diagnostics.Debug.Assert(FEventData.IndexTable != null,"IndexTable is nil");

            for (i = 0; i<FEventData.IndexTable.Count; i++)
            {
 //               indexRecord = (EventData.IndexTable[i] as TTimeSeriesIndexRecord);
                indexRecord = EventData.IndexTable[i];
                indexRecord.Checked = FEventData.InputSettings.ActiveEvents[i];
            }
        }
/************************************************************************
* NAME:    InitChartActiveEventsDefault
* FUNCTIONAL DESCRIPTION: Virtual base class method
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public virtual void InitChartActiveEventsDefault()
        {
            // intentionally empty
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
        }
/************************************************************************
* NAME:    SaveChartActiveEvents
* FUNCTIONAL DESCRIPTION: Save the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
public void SaveChartActiveEvents()
{
  int i;
  TTimeSeriesIndexRecord indexRecord;

  System.Diagnostics.Debug.Assert(FEventData != null,"EventData is nil");
  System.Diagnostics.Debug.Assert(FEventData.IndexTable != null,"IndexTable is nil");

  for (i = 0; i< EventData.IndexTable.Count; i++)
  {
 //     indexRecord = (EventData.IndexTable[i] as TTimeSeriesIndexRecord);
      indexRecord = EventData.IndexTable[i];
      EventData.InputSettings.ActiveEvents[i] = indexRecord.Checked;
  }
}
/************************************************************************
* NAME:   UpdateEventScale
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: not used
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
public virtual void UpdateEventScale()
{
  // intentionally empty
}
/************************************************************************
* NAME:   WhichInterval
* FUNCTIONAL DESCRIPTION:  determine which interval the given value is in
* INPUT PARAMETERS:
*                    iValue : integer   value to test
*                    elimits1 : array of integer   lower bound of each interval
*                    elimits2 : array of integer   upper bound of each interval
* OUTPUT PARAMETERS: none
* RETURN:  result : integer   interval number where value is found (-1 if not in any interval)
************************************************************************/
        public int WhichInterval(int iValue, TLimits[] eLimits ) 
        {
            int i,itest;

            itest = -1;

            for (i = 0; i< eLimits.Length; i++) 
            {
                if((iValue >= eLimits[i].limit1) && (iValue <= eLimits[i].limit2))  // in this interval
                {
                    itest = i;
                    break;
                }
            }
            return itest;
        }
/************************************************************************
* NAME:     EventsInInterval
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: result : boolean   true if events occur, otherwise false
************************************************************************/
        public virtual Boolean EventsInInterval()
        {
            // intentionally blank
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
            return false;
        }
/************************************************************************
* NAME:   FillEventIndexTable
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS:
*                    filter: boolean  use filter controls true/false
* OUTPUT PARAMETERS: none
* RETURN:  result : boolean  true if events found, otherwise false
************************************************************************/
        public virtual Boolean FillEventIndexTable(Boolean filter)
        {
            // intentionally blank
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
            return false;
        }
/************************************************************************
* NAME:     PhaseToEventType
* FUNCTIONAL DESCRIPTION:   Convert phase string to event type
* INPUT PARAMETERS:
*                    start : integer record number where event starts
* OUTPUT PARAMETERS: none
* RETURN:
*                    label : TTimeSeriesEventType  event type
************************************************************************/
        public TTimeSeriesEventType PhaseToEventType(int start)
        {
            const int PHASE_A = 1;
            const int PHASE_C = 3;

            Boolean foundU, foundO, foundI;
            TTimeSeriesEventType eventType;
            string eventPhase;
            int i;
            TTimeSeriesRecord dataRecord;
            GlobalString gs = new GlobalString();

            foundU = false;      // sag
            foundO = false;      // swell
            foundI = false;      // interrupt

            if(start < FEventData.RawDataTable.Count) 
            {
 //               dataRecord = (FEventData.RawDataTable[start] as TTimeSeriesRecord);
                dataRecord = FEventData.RawDataTable[start];
                eventPhase = dataRecord.Phase;

                if(eventPhase == "III") 
                {
                    foundI = true;  // interrupt must occur on all three phases
                }
                else
                {
                    for( i = PHASE_A; i< PHASE_C-1; i++) 
                    {
                        if((eventPhase[i] == 'U') || (eventPhase[i] == 'I')) 
                        {
                            foundU = true;
                        }
                        else if(eventPhase[i] == 'O') 
                        {
                            foundO = true;
                        }
                        if(foundU || foundO)   // use only first event indicator
                        {
                            break;
                        }
                    }
                }

                eventType  = TTimeSeriesEventType.evNone;
                if((foundU && foundO) || (foundU && foundI) || (foundO && foundI)) 
                {
                    eventType = TTimeSeriesEventType.evMulti;   // This will not happen
                }
                else if(foundU) 
                {
                    eventType = TTimeSeriesEventType.evSag;
                }
                else if(foundO) 
                {
                    eventType = TTimeSeriesEventType.evSwell;
                }
                else if(foundI) 
                {
                    eventType = TTimeSeriesEventType.evInterrupt;
                }
            }
            else
            {
                eventType = TTimeSeriesEventType.evNone;
            }

            return eventType;
         }
/************************************************************************
* NAME:     PhaseToEventDescription
* FUNCTIONAL DESCRIPTION:   Create a descriptive event label for event summary based on phase labels
* INPUT PARAMETERS:
*                    aEventType : THMIEventType
* OUTPUT PARAMETERS: none
* RETURN:
*                    result : string  descriptive label for event
************************************************************************/
        public string PhaseToEventDescription(TTimeSeriesEventType aEventType)
        {
            string eventName;

            switch (aEventType)
            {
                case TTimeSeriesEventType.evNone:
                    eventName = "TRIG";
                break;
                case TTimeSeriesEventType.evSag:
                    eventName = "SAG";
                break;
                case TTimeSeriesEventType.evSwell:
                    eventName = "SWELL";
                break;
                case TTimeSeriesEventType.evMulti:
                    eventName = "Multi-Event";
                break;
                case TTimeSeriesEventType.evInterrupt:
                    eventName = "INT";
                break;
                default:
                    eventName = "None";
                    System.Diagnostics.Debug.Assert(false,"Invalid Event Type");
                break;
            }

            return eventName;
        }
/************************************************************************
* NAME:  GetRecordFromEventTime
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: aDateTime : TDateTime   date/time of event
* OUTPUT PARAMETERS: none
* RETURN: result : single  interpolated record number for event
************************************************************************/
public virtual float GetRecordFromEventTime(DateTime aDateTime)
{
  // intentionally empty
  System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
  return 0.0F;
}
/************************************************************************
* NAME:  GetEventRecordFromRecord
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS: myX : double  interpolated record number of event record
* OUTPUT PARAMETERS: none
* RETURN: result : index of Event record
************************************************************************/
public virtual int GetEventRecordFromRecord(double myX)
{
  // intentionally empty
  System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
  return 0;
}
/************************************************************************
* NAME: CBEMAViolation
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS: indexRecord : TTimeSeriesIndexRecord
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesCBEMAViolationType
************************************************************************/
        public virtual TTimeSeriesCBEMAViolationType CBEMAViolation(TTimeSeriesIndexRecord indexRecord)
        {
            // intentionally empty
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
            return TTimeSeriesCBEMAViolationType.cvtNone;
        }
/************************************************************************
* NAME: InitChartTitles
* FUNCTIONAL DESCRIPTION: Set chart titles (main, x and y axes)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public virtual void InitChartTitles()
        {
            if (FChartDisplay.Header.Lines.Length > 1)     // get rid of old titles
            {
                FChartDisplay.Header.Lines = new string[] {"Temp"};
            }
            if(FChartDisplay.Header.Lines.Length > 0) 
            {
                FChartDisplay.Header.Lines[0] = "Click Browse button to read chart data file.";
            }
        }

/******************************************************************************
* MODULE NAME ...  GetChannelPhase
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Determine channel phase from name
*
* INPUT PARAMETERS:
*                    cName : string   channel name
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result : THMIChannelPhase   channel phase
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************/
        public THMIChannelPhase GetChannelPhase(string cName)
        {
            string[] PHASE_A = new string[4] {"a","A","ab","AB"};
            string[] PHASE_B = new string[4] {"b","B","bc","BC"};
            string[] PHASE_C = new string[4] {"c","C","ca","CA"};

            int i;
            THMIChannelPhase myPhase;


            myPhase = THMIChannelPhase.cpOther;

            for (i = 0; i< PHASE_A.Length; i++)
            {
                if(cName.EndsWith(PHASE_A[i]))
                {
                    myPhase = THMIChannelPhase.cpA;
                    break;
                }
            }

            if(myPhase == THMIChannelPhase.cpOther) 
            {
                for (i = 0; i< PHASE_B.Length; i++)
                {
                    if(cName.EndsWith(PHASE_B[i]))
                    {
                        myPhase = THMIChannelPhase.cpB;
                        break;
                    }
                }
            }

            if(myPhase == THMIChannelPhase.cpOther) 
            {
                for (i = 0; i< PHASE_C.Length; i++)
                {
                    if(cName.EndsWith(PHASE_C[i]))
                    {
                        myPhase = THMIChannelPhase.cpC;
                        break;
                    }
                }
            }

            return myPhase;

        }
/******************************************************************************
* MODULE NAME ...  GetChannelType
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Determine channel type from name
*
* INPUT PARAMETERS:
*                    cName : string   channel name
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result: THMIChannelType   channel type
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************/
        public THMIChannelType GetChannelType(string cName)
        {
            if(cName.StartsWith("I")) 
            {
                return THMIChannelType.ctCurrent;
            }
            else if(cName.StartsWith("V")) 
            {
                return THMIChannelType.ctVoltage;
            }
            else if(cName.StartsWith("P")) 
            {
                return THMIChannelType.ctPower;
            }
            else
            {
                return THMIChannelType.ctOther;
            }
        }
/******************************************************************************
* MODULE NAME ...  ChartLineStyle
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Return line style based on channel type
*
* INPUT PARAMETERS:
*                    cType : THMIChannelType   channel type
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result: TPenStyle   line style
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************/
public DashStyle ChartLineStyle(THMIChannelType cType)
{
  const int MAX_TYPE_STYLE = 4;
  DashStyle[] myTypeStyle = new DashStyle[MAX_TYPE_STYLE] { DashStyles.Dash, DashStyles.Solid, DashStyles.DashDot, DashStyles.Solid };

  return myTypeStyle[(int)cType];
}

/******************************************************************************
* MODULE NAME ...  ChartLineColor
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Return line color based on channel phase
*
* INPUT PARAMETERS:
*                    cPhase : THMIChannelPhase   channel phase
*                    index : integer    channel number
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result: TColor   line color
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************/
        public Color ChartLineColor(THMIChannelPhase cPhase, int index)
        {
            const int MAX_COLOR = 12;
            Color[] myColorList = new Color[MAX_COLOR] 
            {
                Colors.Green, Colors.Fuchsia, Colors.Yellow, Colors.Olive, Colors.Teal, Colors.Gray, 
                Colors.Aqua, Colors.LightGray, Colors.DarkGray, Colors.SkyBlue, Colors.MediumSlateBlue, Colors.Maroon
            };

            const int MAX_PHASE_COLOR = 4;
            Color[] myPhaseColor = new Color[MAX_PHASE_COLOR] 
            {
                Colors.Black, Colors.Firebrick, Colors.Blue, Colors.Green
            };

            int adiv, amod, lindex;

            if(cPhase == THMIChannelPhase.cpOther) 
            {
                lindex = index;
                adiv = System.Math.DivRem(lindex,MAX_COLOR,out amod);  // wrap around if not enough colors
                return myColorList[amod];
            }
            else
            {
                return myPhaseColor[Convert.ToInt32(cPhase)]; 
            }
        }
/************************************************************************
* NAME:     EraseChartDisplay
* FUNCTIONAL DESCRIPTION:  Erase content of chart
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void EraseChartDisplay()
        {
            //FChartDisplay.FreeAllSeries();   // throw away previous chart content
            FChartDisplay.Series.Clear();   // throw away previous chart content
            InitChartVariableCount();
            InitChartTitles();
        }
/************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************/
        public virtual void AutoScaleAxes()
        {
            // intentionally empty
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
        }
/************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************/
        public void AutoScaleAxes(out double xMin, out double xMax, out double yMin, out double yMax)
        {
            ChartDisplay.Axes.Bottom.Automatic = false;
            ChartDisplay.Axes.Left.Automatic = false;
            xMin = ChartDisplay.Axes.Bottom.MinXValue;
            xMax = ChartDisplay.Axes.Bottom.MaxXValue;
            yMin = ChartDisplay.Axes.Left.MinYValue;
            yMax = ChartDisplay.Axes.Left.MaxYValue;
            ChartDisplay.Axes.Left.Automatic = false;
        }
/************************************************************************
* NAME:     CBEMAViolationString
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************/
        public virtual string CBEMAViolationString(TTimeSeriesCBEMAViolationType Violation)
        {
            // intentionally empty
            System.Diagnostics.Debug.Assert(false,"Must implement this method in sub-class");
            return " ";
        }
    }
}
