/*****************************************************************************
* UNIT DESCRIPTION:  Base class for LDP charting functions
* SPEC REF: (if used)
* DESIGN REF: (if used)
* 
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************/
using System;
using System.ComponentModel;
using System.Diagnostics;
using Steema.TeeChart;
//#if WPF
using Steema.TeeChart.WPF;
using Steema.TeeChart.WPF.Tools;
using Steema.TeeChart.WPF.Themes;
using System.Windows.Controls;
using Steema.TeeChart.WPF.Styles;
using Microsoft.Win32;
using Steema.TeeChart.WPF.Drawing;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Input;
using System.Windows;
//#endif
using SELGlobalStringMethods;

namespace SELChartClasses
{

    public class TTimeSeriesChartLDP :TTimeSeriesChart
    {
        const string CRLF = "\r\n";     // carriage return line feed
        const string CRLF2 = CRLF + CRLF; 
  
/************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************/
        public  TTimeSeriesChartLDP()
        {
         //
        }
/************************************************************************
* NAME:   GetChartFilename
* FUNCTIONAL DESCRIPTION: Get filename where chart data is saved by default
* INPUT PARAMETERS:  aFileType : TTimeSeriesFileType  (type of file)
*                    ExportPath : default path to file
*                    aID        : ID of meter
* OUTPUT PARAMETERS: none
* RETURN:            filePath : string  path to file where chart data is saved
************************************************************************/
        public override string  GetChartFilename(TTimeSeriesFileType aFileType ,string  ExportPath, string aID)
        {
            string myID;
            string myPath;

            // Choose file names based on Chart type

            myID = aID.Trim();  // If ID is blank, assume it is to be ignored

            if(aFileType == TTimeSeriesFileType.ftCSV)     // CSV file
            {
                if(myID.Length > 0) 
                {
//                    myPath = ExportPath + Format("LDP Export %s_%s.csv", [FormatDateTime("yyyy-mmm-dd", Now),myID]);
                    myPath = ExportPath + "LDP Export " + String.Format("{0:yyyy-mmm-dd}",DateTime.Now) + "_" + myID + ".csv";
                }
                else
                {
//                    myPath = ExportPath + Format("LDP Export %s.csv", [FormatDateTime("yyyy-mmm-dd", Now)]);
                    myPath = ExportPath + "LDP Export " + String.Format("{0:yyyy-mmm-dd}",DateTime.Now) + ".csv";
                }
            }
            else
            {                           // ASCII encoded binary file
                if(myID.Length > 0)
                {
//                    myPath = ExportPath + Format("LDP Export %s_%s.txt", [FormatDateTime("yyyy-mmm-dd", Now),myID]);
                    myPath = ExportPath + "LDP Export " + String.Format("{0:yyyy-mmm-dd}",DateTime.Now) + "_" + myID + ".txt";
                }
                else
                {
//                    myPath = ExportPath + Format("LDP Export %s.txt", [FormatDateTime("yyyy-mmm-dd", Now)]);
                    myPath = ExportPath + "LDP Export " + String.Format("{0:yyyy-mmm-dd}",DateTime.Now) + ".txt";
                }
            }
            return myPath;
        }
/************************************************************************
* NAME:  InitChart
* FUNCTIONAL DESCRIPTION: Initialize chart class members before importing
*                         LDP report file (column convention must be followed)
* INPUT PARAMETERS: fileType : TTimeSeriesFileType   type of file (ftReport,...)
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public override void InitChart(TTimeSeriesFileType fileType )
        {
            int[] standardLDP=new int[5] {-1,1,2,-1,3};   // record, date, time, phase, status
            int i,lastStandard;

            System.Diagnostics.Debug.Assert(EventData!=null,"EventData is null");
            base.InitChart(fileType);

            EventData.ChartType = TTimeSeriesChartType.ctLDP;
            EventData.ChartName = "LDP";
            EventData.SetLengthHeaderColumns(standardLDP.Length);
            EventData.SetLengthDataColumns(EventData.RawDataLabels.Count);
            lastStandard = 0;  // find last standard column used

            // Initialize standard columns.  (Independent variables)
            for( i = 0; i<standardLDP.Length; i++)
            {
                EventData.HeaderColumns[i] = standardLDP[i];
                lastStandard = System.Math.Max(lastStandard,standardLDP[i]);  // find last column used
            }
            // Initialize data columns (dependent variables)
            for( i = 0; i<EventData.RawDataLabels.Count; i++)
            {
                EventData.DataColumns[i] = lastStandard + i + 1;
            }
        }
/************************************************************************
* NAME:    InitChartVariableCount
* FUNCTIONAL DESCRIPTION:
* Set the chart variable count based on chart type and event count
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void InitChartVariableCount()
        {

            System.Diagnostics.Debug.Assert(EventData!=null,"EventData is null");

            base.InitChartVariableCount();

            if(EventData.IndexTable.Count > 0)    // LDP SER events are active
            {
                ChartHasEvents = true;
            }
        }
/************************************************************************
* NAME:    SetChartActiveChannels
* FUNCTIONAL DESCRIPTION:  Set the active series in the chart based on
*                          the user"s input selections.
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void InitChartActiveChannels()
        {
            int i,sc, ac;
            Boolean status;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");

            sc = ChartDisplay.Series.Count;
            ac = EventData.InputSettings.ActiveChannels.Length;

            System.Diagnostics.Debug.Assert(sc <= ac, "Array size mismatch");

            // update chart active channels

            for( i = 0; i<ChartDisplay.Series.Count; i++)
            {
                status =  EventData.InputSettings.ActiveChannels[i];
                ChartDisplay.Series[i].Active = status;
            }

        }
/************************************************************************
* NAME:    InitChartActiveChannelsDefault
* FUNCTIONAL DESCRIPTION:
* Set the default state of active channels
* INPUT PARAMETERS:    reset : boolean  reset channel status to default if true
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void InitChartActiveChannelsDefault(Boolean reset)
        {
            int i,oldCount,newCount;
            Boolean[] oldState;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");

            /*The Active Channel array is a list of booleans that indicates whether
            each channel is selected for display in the chart (active).  There may
            be one additional value in the chart to represent event markers.  These
            markers indicate the occurrence of an LDP event during the time interval.
            If events are present, one additional member is added to the channel array.
            */
            oldCount = EventData.InputSettings.ActiveChannels.Length;
            newCount = ChartVariableCount;

            if(ChartHasEvents)  // allocate one more position and disable events by default
            {
                newCount = ChartVariableCount+1;
            }

            if(reset || (newCount != oldCount))    // length has changed
            {
                if (!reset)      // save active channel state before re-allocating
                {
                    oldState = new Boolean[oldCount];
                    for (i = 0; i < oldCount; i++)  // remember previous active channel state
                    {
                        oldState[i] = EventData.InputSettings.ActiveChannels[i];
                    }

                    InitActiveChannels(newCount);

                    oldCount = System.Math.Min(oldCount, newCount);
                    for (i = 0; i < oldCount; i++)  // restore channel states
                    {
                        EventData.InputSettings.ActiveChannels[i] = oldState[i];
                    }
                }
                else
                {
                    InitActiveChannels(newCount);
                }
            }
        }
        /************************************************************************
        * NAME:    InitActiveChannels
        * FUNCTIONAL DESCRIPTION:  initialize active channel array
        * Save the active series in the chart
         INPUT PARAMETERS:    newCount : int  new number of active channels
        * OUTPUT PARAMETERS:   none
        * RETURN:              none
        ************************************************************************/

        public void InitActiveChannels(int newCount)
        {
            int i;

            EventData.InputSettings.SetActiveChannelLength(newCount);
            EventData.InputSettings.ActiveChannels[newCount - 1] = false; // disable last channel (event)

            for (i = 0; i < ChartVariableCount; i++)
            {
                EventData.InputSettings.ActiveChannels[i] = true;
            }
        }
/************************************************************************
* NAME:    SaveChartActiveChannels
* FUNCTIONAL DESCRIPTION:
* Save the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void SaveChartActiveChannels()
        {
            int i;
            Boolean status;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");
            System.Diagnostics.Debug.Assert(EventData.InputSettings != null,"InputSettings is null");

            for (i = 0; i< ChartDisplay.Series.Count; i++)
            {
                status = ChartDisplay.Series[i].Active;
                EventData.InputSettings.ActiveChannels[i] = status;
            }
        }
/************************************************************************
* NAME:   UpdateEventScale
* FUNCTIONAL DESCRIPTION:  update yvalue of SER event markers
* INPUT PARAMETERS: not used
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public override void UpdateEventScale()
        {
            //  PointSeries myPointSeries;
            Series myPointSeries;
            float yValue;
            int i;

            //  inherited UpdateEventScale;

            /*LDP records are charted using the record number as the independent variable.
            If there are event records, they are represented as an additional dataset
            and displayed as points (point series).  Events are recorded independently
            from the LDP data, so their timestamp may not coincide with any particular
            LDP timestamped record.  Therefore, it is necessary to obtain an approximate
            (interpolated) record number corresponding for the event record so that
            it may be displayed with the LDP data. Because the event record has no
            magnitude (y-value), an artificial value is obtained by finding the minimum
            value of the LDP data presently displayed, and assigning this value to
            each event record.  This way, the event markers will always be visible
            near the bottom of the chart, no matter what the range of LDP magnitudes.
            */
       //     if ((ChartDisplay.Series[ChartDisplay.Series.Count - 1] is PointSeries) &&
            if((ChartDisplay.Series[ChartDisplay.Series.Count -1] is Steema.TeeChart.WPF.Styles.Points) &&
                ChartDisplay.Series[ChartDisplay.Series.Count - 1].Active)    // find manimum of LDP values when events excluded
            {
                ChartDisplay.Series[ChartDisplay.Series.Count -1].Active = false;
                AutoScaleAxes();   // set scale without events

                myPointSeries = ChartDisplay.Series[ChartDisplay.Series.Count-1];
                yValue = Convert.ToSingle(ChartDisplay.Axes.Left.MinYValue);
                for (i = 0; i< myPointSeries.Count; i++) 
                {
                    myPointSeries.YValues[i] = yValue;
                }

                ChartDisplay.Series[ChartDisplay.Series.Count -1].Active = true;
            }
        }
/************************************************************************
* NAME:  FillEventIndexTable
* FUNCTIONAL DESCRIPTION:  Scan LDP data, identify events, create index of events
* INPUT PARAMETERS:
*                    filter: boolean  use filter controls true/false
* OUTPUT PARAMETERS: none
* RETURN:
*                    found : boolean  true if events found, otherwise false
************************************************************************/
        public override Boolean FillEventIndexTable(Boolean filter)
        {
            DateTime userStartDate, userEndDate, recordDate, ldpStart, ldpEnd;
            TTimeSeriesIndexRecord indexRecord;
            Boolean EventBeginChecked, EventEndChecked;
            int i;
            TTimeSeriesLDPEventRecord dataRecord;
            TTimeSeriesRecord ldpRecord;
            TTimeSeriesInputSettingsLDP myLDP;
            GlobalString gs = new GlobalString();

            /*Create a list of LDP events that occur within the time interval of LDP records.
            LDP events (SER) are recorded independently from LDP records and therefore may not
            have coincident time stamps.  Furthermore, when LDP data is obtained from the
            device, all available SER records are retrieved (up to 512 most recent), so
            it is possible that none of these events occurred during the interval that
            LDP records were saved.  This function determines which (if any) SER events
            occurred during the interval when LDP records were saved.
            */
            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");
            System.Diagnostics.Debug.Assert(EventData.LDPEventTable != null,"LDPEventTable is null");

            EventData.IndexTable.Clear();   // erase old index

            if((EventData.LDPEventTable.Count == 0) ||
                (EventData.RawDataTable.Count == 0)) 
            {
                return false;  // no data
            }

            // get users supplied start/end dates in case they are active

            // if date/time is "Checked", the user has enabled filtering
            myLDP = (TTimeSeriesInputSettingsLDP) EventData.InputSettings;
            EventBeginChecked = myLDP.BeginDateActive;
            EventEndChecked = myLDP.EndDateActive;
            // Users date and time interval
            userStartDate = myLDP.EndDateTime;
            userEndDate = myLDP.EndDateTime;

            // get date/time of first and last LDP records

            ldpRecord = EventData.RawDataTable[0];
            ldpStart = ldpRecord.DateTime;   // starting date/time of LDP data records
            ldpRecord = EventData.RawDataTable[EventData.RawDataTable.Count-1];
            ldpEnd= ldpRecord.DateTime;      // ending date/time of LDP data records

            // see if any events occurred during LDP interval

            dataRecord = EventData.LDPEventTable[0];
            recordDate = dataRecord.DateTime;   // starting date/time of event records
            if(recordDate > ldpEnd)    // events all after LDP interval
            {
                return false;
            }
            dataRecord = EventData.LDPEventTable[EventData.LDPEventTable.Count-1];
            recordDate= dataRecord.DateTime;      // ending date/time of event records
            if(recordDate < ldpStart)    // events all before LDP interval
            {
                return false;
            }

            // scan LDP event records searching for events that are within the LDP interval

            for (i = 0; i<EventData.LDPEventTable.Count; i++) 
            {
                dataRecord = EventData.LDPEventTable[i];
                recordDate = dataRecord.DateTime;

                if(recordDate < ldpStart)      // LDP start
                {
                    continue;
                }
                else if(recordDate > ldpEnd)       // LDP end
                {
                    continue;
                }
                else if (filter && EventEndChecked && (gs.CompareDateTime(recordDate,userStartDate) < 0)) 
                {
                    continue;  // before { of user"s filter interval
                }
                else if (filter && EventEndChecked && (gs.CompareDateTime(recordDate,userEndDate) > 0)) 
                {
                    break;  // past end of user"s filter interval
                }

                // Test for type of event

                indexRecord = new TTimeSeriesIndexRecord();  // new index record
                indexRecord.DateTime = dataRecord.DateTime;
                indexRecord.EventName = dataRecord.EventName;
                indexRecord.Description = dataRecord.Description;

                EventData.IndexTable.Add(indexRecord);  // add new record into event index
            }

            return true; // if we got here, it is ok to generate event list and chart
        }
/************************************************************************
* NAME:     EventsInInterval
* FUNCTIONAL DESCRIPTION:  Determine if events occur during data interval
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: result : boolean   true if events occur, otherwise false
************************************************************************/
        public override Boolean EventsInInterval()
        {
            TTimeSeriesRecord dataRecord;
            TTimeSeriesIndexRecord indexRecord;

            if((EventData.IndexTable.Count <= 0) || (EventData.RawDataTable.Count <= 0)) 
            {
                return false;
            }

            dataRecord = EventData.RawDataTable[EventData.RawDataTable.Count-1];
            indexRecord = EventData.IndexTable[0];
            if(indexRecord.DateTime > dataRecord.DateTime)  // first event after last data record
            {
                return false;
            }

            dataRecord = EventData.RawDataTable[0];
            indexRecord = EventData.IndexTable[EventData.IndexTable.Count-1];
            if(indexRecord.DateTime < dataRecord.DateTime)  // last event before first data record
            {
                return false;
            }

            return true;      // there are some events in the data interval

        }
/************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Reset chart axis limits to leave small interval around limits of data
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************/
        public override void AutoScaleAxes()
        {
            double xMin, xMax, yMin, yMax;

            /*  TeeChart autoscaling causes an exact fit to the data. Since this sometimes not
            *  visually appealing, this scaling function subtracts 5% from the limits of
            *  the horizontal and vertical axes so that chart is clearly visible in the chart
            *  area.*/

            base.AutoScaleAxes(out xMin,out xMax,out yMin,out yMax);

            ChartDisplay.Axes.Bottom.SetMinMax(xMin - (xMax - xMin) / 20.0, xMax + (xMax - xMin) / 20.0);
            ChartDisplay.Axes.Left.SetMinMax(yMin - (yMax - yMin) / 20.0, yMax + (yMax - yMin) / 20.0);
        }
/************************************************************************
* NAME:  GetRecordFromEventTime
* FUNCTIONAL DESCRIPTION:  return interpolated record number for event
* INPUT PARAMETERS: aDateTime : TDateTime   date/time of event
* OUTPUT PARAMETERS: none
* RETURN: result : single  interpolated record number for event
************************************************************************/
        public override float GetRecordFromEventTime(DateTime aDateTime)
        {
            int imin,imax,ipos,rec1,rec2;
            float fraction;
            Boolean more;
            DateTime myDateTime1, myDateTime2;
        
            /*This method is used to assign an artifical record number to an event
            with the given timestamp.  Find the pair of LDP records that bracket
            the given timestamp and assign an interpolated record number to represent
            the position of the event record between these records.
            */
            imin = 0;
            if (aDateTime < EventData.RawDataTable[imin].DateTime)
                {
                return -1.0F;  // out of range
            }
            imax = EventData.RawDataTable.Count-1;
            if(aDateTime > EventData.RawDataTable[imax].DateTime) 
            {
                return -1.0F;  // out of range
            }

            more = true;
            imin = 0;
            imax = EventData.RawDataTable.Count-1;

            // binary search for best position

            while (more) 
            {
                ipos = (imin + imax) / 2;
                if(aDateTime > EventData.RawDataTable[ipos].DateTime)     // in top half
                {
                    imin = ipos;
                }
                else                                     // in bottom half
                {
                    imax = ipos;
                }
                if((imax - imin) < 2) 
                {
                    more = false;
                }
            }
            myDateTime1 = EventData.RawDataTable[imin].DateTime;
            rec1 = EventData.RawDataTable[imin].RecordNumber-1;
            myDateTime2 = EventData.RawDataTable[imax].DateTime;
            rec2 = EventData.RawDataTable[imax].RecordNumber-1;

            // get interpolated record number between bounding timestamped records

            fraction = Convert.ToSingle(aDateTime.Subtract(myDateTime1).TotalDays / myDateTime2.Subtract(myDateTime1).TotalDays);
            //fraction = Convert.ToSingle(aDateTime - myDateTime1) / Convert.ToSingle(myDateTime2 - myDateTime1);
//            fraction = (aDateTime - myDateTime1) / (myDateTime2 - myDateTime1);

            return (rec1 + fraction * (rec2 - rec1));
        }
/************************************************************************
* NAME:  GetEventRecordFromRecord
* FUNCTIONAL DESCRIPTION: find LDP record date/time that closely matches record number
* INPUT PARAMETERS: myX : double  interpolated record number of event record
* OUTPUT PARAMETERS: none
* RETURN: result : index of Event record
************************************************************************/
        public override int GetEventRecordFromRecord(double myX)
        {
            int rec1, rec2;
            DateTime dateTime1, dateTime2, myDateTime;
            int imin, imax, ipos;
            Boolean more;
            TTimeSeriesIndexRecord indexRecord;
            GlobalString gs = new GlobalString();

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");

            /*LDP records are charted using the record number as the independent variable.
            If there are event records, they are represented as an additional dataset
            and displayed as points (point series).  Events are recorded independently
            from the LDP data, so their timestamp may not coincide with any particular
            LDP timestamped record.  In order to chart events with LDP data, the
            event records are assigned an artifical record number that is an interpolation
            between the LDP record numbers that lie immediately before and after the
            timestamp of the event record.

            This method is used to find the Event record corresponding to the
            event marker where the mouse cursor is pointing.  It is used when
            doing an interactive data query.
            */
            if((EventData.RawDataTable.Count == 0) || (EventData.IndexTable.Count == 0)) 
            {
                return 0;  // This should never happen because the chart could not exist
            }


            rec1 = Convert.ToInt32(myX);  // get LDP record number before and after event
            rec2 = rec1 + 1;
            dateTime1  = EventData.RawDataTable[rec1].DateTime;
            dateTime2 = EventData.RawDataTable[rec2].DateTime;
            myDateTime = gs.InterpolateDateTime(dateTime1, dateTime2, myX);

            more = true;
            imin = 0;
            imax = EventData.IndexTable.Count-1;

            // binary search to find event record with closest date/time to chosen record

            while (more)
            {
                ipos = (imin + imax) / 2;
                indexRecord = EventData.IndexTable[ipos];
                if(myDateTime > indexRecord.DateTime)      // in top half
                {
                    imin = ipos;
                }
                else                                     // in bottom half
                {
                    imax = ipos;
                }

                if((imax - imin) < 2) 
                {
                    more = false;
                }
            }

            // We now know the interval that contains the event record we are looking for
            if(imin == imax) 
            {
                return imin;
            }
            else
            {
                dateTime1 = EventData.IndexTable[imin].DateTime;
                dateTime2 = EventData.IndexTable[imax].DateTime;
                if((myDateTime - dateTime1) < (dateTime2 - myDateTime)) 
                {
                    return imin;   // lower limit
                }
                else
                {
                    return imax;   // upper limit
                }
            }
        }
/************************************************************************
* NAME: InitChartTitles
* FUNCTIONAL DESCRIPTION: Set chart titles (main, x and y axes)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public override void InitChartTitles()
        {
            //string title;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");
            base.InitChartTitles();   // allocates title array

            if(EventData.RawDataTable.Count > 1) 
            {
                //title = EventData.MeterID + " Event(s): ";

                ChartDisplay.Header.Lines[0] = EventData.MeterID + "- Load Profile History";
                ChartDisplay.Header.AdjustFrame = true;
                ChartDisplay.Axes.Left.Title.Caption = "Channel Value";

                ChartDisplay.Axes.Left.Title.Angle = 90;    // vertical orientation

                if(EventData.InputSettings.ChartAxisDateTime)     // display date/time X axis labels
                {
                    ChartDisplay.Axes.Bottom.Title.Caption = "Time";
                }
                else
                {
                    ChartDisplay.Axes.Bottom.Title.Caption = "Record Number";
                }
            }
        }
/************************************************************************
* NAME:    AppendCSVHeader
* FUNCTIONAL DESCRIPTION:  Append header records for CSV file to string
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: MeterBuffer : string  file content
* RETURN: result : boolean   status of append (true = success)
************************************************************************/
        public override Boolean AppendCSVHeader(ref string MeterBuffer)
        {
            int i;
            TTimeSeriesLDPEventRecord dataRecord;
            TTimeSeriesLDPChannelRecord channelRecord;
            string myDate, myTime;
            Boolean result;
            GlobalString gs = new GlobalString();

            result = base.AppendCSVHeader(ref MeterBuffer);

            // Add Channel Data

            MeterBuffer = MeterBuffer +
            "ChannelName, ChannelFunction, ChannelRule, ChannelUnits, ChannelScale" + CRLF;
            for (i = 0; i< EventData.LDPChannelTable.Count; i++) 
            {
                channelRecord = EventData.LDPChannelTable[i];
                MeterBuffer = MeterBuffer +
                channelRecord.ChannelName + ", " +
                channelRecord.ChannelFunction + ", " +
                channelRecord.ChannelRules + ", " +
                channelRecord.ChannelUnits + ", " +
                gs.NonLocaleFloatToStr(channelRecord.ChannelScale) + CRLF;
            }

            // Add SER Event Data

            MeterBuffer = MeterBuffer + "EventDate, EventTime, EventCode, EventDescription" + CRLF;
            for( i = 0; i< EventData.LDPEventTable.Count; i++) 
            {
                dataRecord = EventData.LDPEventTable[i]; // get the event index record
                myDate = String.Format("{0:MM/dd/yyyy}",dataRecord.DateTime);  // localization date
                myTime = String.Format("{0:hh:mm:ss.fff}",dataRecord.DateTime);
                MeterBuffer = MeterBuffer + myDate + ", " + myTime + ", " +
                           dataRecord.EventCode + ", \"" + dataRecord.Description + "\"" + CRLF;
            }
            return result;
        }
/************************************************************************
* NAME:    AppendCSVData
* FUNCTIONAL DESCRIPTION: Append CSV data from HMI data stored
*                         in memory resident TimeSeries structure
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      MeterBuffer : string  file content
* RETURN:                 result : boolean   true if data appended, otherwise false
************************************************************************/
        public override Boolean AppendCSVData(ref string MeterBuffer)
        {
            int i,j,channelCount,recordCount,pos;      // indices
            string myDateTime,myValue;
            TTimeSeriesRecord dataRecord;
            double rValue;
            GlobalString gs = new GlobalString();

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");
            System.Diagnostics.Debug.Assert(EventData.RawDataLabels != null,"RawDataLabels is null");
            System.Diagnostics.Debug.Assert(EventData.RawDataTable != null,"RawDataTable is null");

            channelCount = EventData.RawDataLabels.Count;
            recordCount = EventData.RawDataTable.Count;

            if((channelCount == 0) || (recordCount == 0))    // no data available
            {
                MeterBuffer = MeterBuffer + "# No LDP recorder data available." + CRLF;
                return false;
            }
            else
            {
                MeterBuffer = MeterBuffer + EventData.ChartField + ", Date, Time, Status, ";

                // export channel names
                for (i = 0; i< channelCount; i++)
                {
                    MeterBuffer = MeterBuffer + EventData.RawDataLabels[i];
                    if(i < channelCount-1) 
                    {
                        MeterBuffer = MeterBuffer + ", ";
                    }
                }

                MeterBuffer = MeterBuffer + CRLF;

                // export channel data
                for (i = 0; i<recordCount; i++)
                {
                    dataRecord = EventData.RawDataTable[i];
                    MeterBuffer = MeterBuffer + String.Format("{0:d}",dataRecord.RecordNumber) + ", ";
                    myDateTime = String.Format("{0:MM/dd/yyyy, HH:mm:sss.fff}",dataRecord.DateTime);
                    //DateTimeToString(myDateTime ,"MM/dd/yyyy,  HH:mm:sss.fff",dataRecord.DateTime);
                    MeterBuffer = MeterBuffer + myDateTime + ", \"" + dataRecord.Status + "\"";

                    for( j = 0; j<channelCount; j++)
                    {
                        rValue = dataRecord.Value[j] + 0.005;  // account for 2 decimal digit precision error
                        myValue = gs.NonLocaleFloatToStr(Convert.ToSingle(rValue)); // may or may not have decimal pt.
                        pos = myValue.IndexOf(".");  // look for decimal
                        if(pos > 0) 
                        {
                            myValue = myValue.Substring(0,pos+2);   // keep only 2 decimal places
                        }
                        MeterBuffer = MeterBuffer + ", " + myValue;
                    }

                    MeterBuffer = MeterBuffer + CRLF;
                }
                return true;
            }
        }
/************************************************************************
* NAME:    AppendReport
* FUNCTIONAL DESCRIPTION: Append LDP summary to string
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   MeterBuffer : string  file content
* RETURN:              result : boolean   true if summary appended, otherwise false
************************************************************************/
        public override Boolean AppendReport(out string MeterBuffer)
        {
            const string LDP_STATUS =
                    "STATUS CODES:" + CRLF +
                    "1 - Daylight savings time is in effect during or at start of interval" + CRLF +
                    "2 - Power fail within interval (missing data)" + CRLF +
                    "3 - Clock reset forward during interval" + CRLF +
                    "4 - Clock reset backwards during interval" + CRLF +
                    "5 - Skipped interval" + CRLF +
                    "6 - TEST mode data" + CRLF +
                    "7 - Data overwrite";

            int i,j;      // indices
            TTimeSeriesRecord dataRecord;
            string myDataFormat, myFunction, myDate, myTime;
            TTimeSeriesLDPChannelRecord channelRecord;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings();

            EventData.InitDateTimeFormat(TDateFormatType.ftInternational, ref myFormat);
            gs.DateFormat2DigitYear(ref myFormat);
            myDataFormat = " {0,10:0.00}";
            MeterBuffer = " ";

            if(EventData.DataColumnArray.Length > 0) 
            {
                if(EventData.LDPChannelTable.Count > 0)   // see if there is channel information
                {
                    channelRecord = EventData.LDPChannelTable[0];
                    myFunction = channelRecord.ChannelFunction;
                }
                else
                {
                    myFunction = "N/A";
                }

                // Write report header information
                myDate = gs.DateToStr(EventData.ChartDateTime,myFormat);  // use setting date format
                myTime = string.Format("{0:hh:mm:ss}",EventData.ChartDateTime);

                MeterBuffer = string.Format("{0,-60}",EventData.MeterID) +
                                " Date: " + myDate +
                                " Time: " + myTime + CRLF;
                MeterBuffer = MeterBuffer + string.Format("{0,-60}",EventData.MeterTID) + " Time Source: " + 
                              EventData.TimeSource + CRLF2;
                MeterBuffer = MeterBuffer + string.Format("{0,-60}",EventData.MeterFID) + " Function: " + 
                              myFunction + CRLF2;
                MeterBuffer = MeterBuffer + "#    DATE     TIME  ";

                if(EventData.DataColumns.Length > 0)         // write channel names
                {
                    for (i = 0; i<EventData.DataColumns.Length; i++)
                    {
                        MeterBuffer = MeterBuffer + string.Format("{0,11}",EventData.RawDataLabels[i].Trim());
                    }
                    MeterBuffer = MeterBuffer + "   STATUS" + CRLF;
                }
                else
                {
                    MeterBuffer = MeterBuffer + "No LDP records found." + CRLF;
                }

                // Write table of LDP data

                for (i = 0; i<EventData.RawDataTable.Count; i++)
                {
                    dataRecord = EventData.RawDataTable[i];
                    MeterBuffer = MeterBuffer + string.Format("{0:d}",dataRecord.RecordNumber) + "  ";
                    myDate = gs.DateToStr(dataRecord.DateTime,myFormat);  // use setting date format
                    myTime = string.Format("{0:hh:mm:ss}",dataRecord.DateTime);
                    MeterBuffer = MeterBuffer + myDate + " " + myTime + "  ";

                    for (j = 0; j<EventData.DataColumns.Length; j++) // write channel values
                    {
                        MeterBuffer = MeterBuffer + string.Format(myDataFormat,dataRecord.Value[j]);
                    }
                    if(dataRecord.Status.Trim().Length == 0) 
                    {
                        dataRecord.Status = "...";
                    }
                    MeterBuffer = MeterBuffer + "  " + dataRecord.Status + CRLF;
                }
                MeterBuffer = MeterBuffer + CRLF + LDP_STATUS;  // append status codes
                return true;
            }
            else
            {
                return false;     // no data to write
            }
        }
    }
}
