/*****************************************************************************
* UNIT DESCRIPTION:  Base class for SSI charting functions
* SPEC REF: (if used)
* DESIGN REF: (if used)
* 
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************/
using System;
using System.ComponentModel;
using System.Diagnostics;
using Steema.TeeChart;
//#if WPF
using Steema.TeeChart.WPF;
using Steema.TeeChart.WPF.Tools;
using Steema.TeeChart.WPF.Themes;
using System.Windows.Controls;
using Steema.TeeChart.WPF.Styles;
using Microsoft.Win32;
using Steema.TeeChart.WPF.Drawing;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Input;
using System.Windows;
//#endif
using SELGlobalStringMethods;

namespace SELChartClasses
{

    public class TTimeSeriesChartSSI:TTimeSeriesChart
    {
        const string CRLF = "\r\n";     // carriage return line feed
        const int PHASE_A = 0;
        const int PHASE_B = 1;
        const int PHASE_C = 2;
        const string INTERRUPT_3P = "III";
        const string NOMINAL_3P = "...";
        const char SAG_EVENT = 'U';
        const char SWELL_EVENT = 'O';
        const char INTERRUPT_EVENT = 'I';
        const char NOMINAL_EVENT = '.';
        const string PRE_DISTURBANCE = "P";
        const string POST_DISTURBANCE = "E";
        const string INVALID_RECORD = "X";
        const string FAST_RECORD = "F";
        const string READY_RECORD = "R";

/************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************/
        public TTimeSeriesChartSSI()
        {
            //  inherited Create(aowner);
        }
/************************************************************************
* NAME:   GetChartFilename
* FUNCTIONAL DESCRIPTION: Get filename where chart data is saved by default
* INPUT PARAMETERS:  aFileType : TTimeSeriesFileType  (type of file)
*                    ExportPath : default path to file
*                    aID        : ID of meter
* OUTPUT PARAMETERS: none
* RETURN:            filePath : string  path to file that was saved
************************************************************************/
        public override string GetChartFilename(TTimeSeriesFileType aFileType, string ExportPath, string aID)
        {
            string myID;
            string myPath;

            myID = aID.Trim();  // If ID is blank, assume it is to be ignored

            // Choose file names based on Chart type

            if(aFileType == TTimeSeriesFileType.ftCSV) 
            {
                if(myID.Length > 0) 
                {
                //      myPath = ExportPath + Format("SSI Export %s_%s.csv", [FormatDateTime("yyyy-mmm-dd", Now),myID]);
                    myPath = ExportPath + "SSI Export " + string.Format("{0:yyyy-mmm-dd}",DateTime.Now) + "_" + myID + ".csv";
                }
                else
                {
                //myPath = ExportPath + Format("SSI Export %s.csv", [FormatDateTime("yyyy-mmm-dd", Now)]);
                    myPath = ExportPath + "SSI Export " + string.Format("{0:yyyy-mmm-dd}",DateTime.Now) + ".csv";
                }
            }
            else
            {
                if(myID.Length > 0) 
                {
                    myPath = ExportPath + "SSI Export " + string.Format("{0:yyyy-mmm-dd}",DateTime.Now) + "_" + myID + ".txt";
                //      myPath = ExportPath + Format("SSI Export %s_%s.txt", [FormatDateTime("yyyy-mmm-dd", Now),myID]);
                }
                else
                {
                    myPath = ExportPath + "SSI Export " + string.Format("{0:yyyy-mmm-dd}",DateTime.Now) + ".txt";
                //     myPath = ExportPath + Format("SSI Export %s.txt", [FormatDateTime("yyyy-mmm-dd", Now)]);
                }
            }
            return myPath;
        }
/************************************************************************
* NAME:  InitChart
* FUNCTIONAL DESCRIPTION: Initialize chart class members before importing SSI report file
* INPUT PARAMETERS: aFileType : TTimeSeriesFileType  file type (ftReport...)
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public override void InitChart(TTimeSeriesFileType aFileType)
        {
            int[] standardSSI_REP = new int[5] {-1,1,2,12,13};   // record, date, time, phase, status
            int[] standardSSI_CSV = new int[5] {-1,1,2,11,12};   // record, date, time, phase, status
            int[] standardSSI_CSV_4XX =new int[5] {-1,1,2,6,7};   // record, date, time, phase, status
            const int SIZE_COLUMNS_4XX = 3;    // only 3 voltages (Va, Vb, Vc)
            const int SIZE_COLUMNS = 8;       // Currents and Voltages (Ia,Ib,Ic,Ig,In,Va,Vb,Vc)
            const int FIRST_COLUMN = 3;       // First current or voltage

            int i,size;

            base.InitChart(aFileType);
            EventData.ChartType = TTimeSeriesChartType.ctSSI;
            EventData.ChartName = "VSSI";

            if(aFileType == TTimeSeriesFileType.ftReport) 
            {
                EventData.SetLengthHeaderColumns(standardSSI_REP.Length);
                EventData.SetLengthDataColumns(SIZE_COLUMNS);
                // Initialize standard columns
                for (i = 0; i<standardSSI_REP.Length; i++)
                {
                    EventData.HeaderColumns[i] = standardSSI_REP[i];
                }
                size = SIZE_COLUMNS;
            }
            else
            {
                if(EventData.DeviceType == TTimeSeriesDeviceType.dtSEL4XX) 
                {
                    EventData.SetLengthHeaderColumns(standardSSI_CSV_4XX.Length);
                    EventData.SetLengthDataColumns(SIZE_COLUMNS_4XX);
                    // Initialize standard columns
                    for (i = 0; i< standardSSI_CSV_4XX.Length; i++)
                    {
                        EventData.HeaderColumns[i] = standardSSI_CSV_4XX[i];
                    }
                    size = SIZE_COLUMNS_4XX;
                }
                else
                {
                    EventData.SetLengthHeaderColumns(standardSSI_CSV.Length);
                    EventData.SetLengthDataColumns(SIZE_COLUMNS);
                    // Initialize standard columns
                    for ( i = 0; i<standardSSI_CSV.Length; i++) 
                    {
                        EventData.HeaderColumns[i] = standardSSI_CSV[i];
                    }
                    size = SIZE_COLUMNS;
                }
            }
            // Initialize data columns
            for (i = 0; i< size; i++)
            {
                EventData.DataColumns[i] = FIRST_COLUMN + i;
            }
        }
/************************************************************************
* NAME:    InitChartVariableCount
* FUNCTIONAL DESCRIPTION:
* Set the chart variable count based on chart type and event count
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void InitChartVariableCount()
        {

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");

            base.InitChartVariableCount();  // initializes ChartHasEvents

            // update chart variable count to include CBEMA curves if appropriate

            if(EventData.IndexTable.Count > 0)      // SSI CBEMA curves are active
            {
                ChartHasEvents = true;
            }
        }
/************************************************************************
* NAME:    InitChartActiveChannels
* FUNCTIONAL DESCRIPTION:
* Set the active series in the chart
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void InitChartActiveChannels()
        {
            int i,sc,ac;
            Boolean status;
            TTimeSeriesInputSettingsSSI mySSI;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");

            ac = EventData.InputSettings.ActiveChannels.Length;
            sc = ChartDisplay.Series.Count;
            mySSI = (TTimeSeriesInputSettingsSSI)EventData.InputSettings;

            System.Diagnostics.Debug.Assert(sc <= ac, "Array size mismatch");

            // update chart active channels

            if((EventData.ChartType == TTimeSeriesChartType.ctSSI) && mySSI.ChartCBEMA) 
            {
                return;
            }

            for (i = 0; i<ChartDisplay.Series.Count; i++)
            {
                status =  EventData.InputSettings.ActiveChannels[i];
                ChartDisplay.Series[i].Active = status;
            }
        }
/************************************************************************
* NAME:    InitChartActiveChannelsDefault
* FUNCTIONAL DESCRIPTION:
** Set the default state of active channels
 INPUT PARAMETERS:    reset : boolean  reset channel status to default if true
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void InitChartActiveChannelsDefault(Boolean reset)
        {
            int i;
            int oldCount, newCount;
            Boolean[] oldState;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");

            oldCount = EventData.InputSettings.ActiveChannels.Length;
            newCount = ChartVariableCount;

            if(ChartHasEvents)  // allocate one more position and disable events by default
            {
                newCount = ChartVariableCount+1;
            }

            if(reset || (newCount != oldCount))    // length has changed
            {
                if (!reset)      // save active channel state before re-allocating
                {
                    oldState = new Boolean[oldCount];
                    for (i = 0; i < oldCount; i++)   // remember previous active channel state
                    {
                        oldState[i] = EventData.InputSettings.ActiveChannels[i];
                    }

                    InitActiveChannels(newCount);

                    // restore old state up to current length of array

                    oldCount = System.Math.Min(oldCount, newCount);
                    for (i = 0; i < oldCount; i++)  // restore channel states
                    {
                        EventData.InputSettings.ActiveChannels[i] = oldState[i];
                    }
                }
                else
                {
                    InitActiveChannels(newCount);
                }
            }
        }
/************************************************************************
* NAME:    SaveChartActiveChannels
* FUNCTIONAL DESCRIPTION:
* Save the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void SaveChartActiveChannels()
        {
            int i;
            TTimeSeriesInputSettingsSSI mySSI;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");
            System.Diagnostics.Debug.Assert(EventData.InputSettings != null,"InputSettings is null");

            mySSI = (TTimeSeriesInputSettingsSSI)EventData.InputSettings;

            if((EventData.ChartType ==  TTimeSeriesChartType.ctSSI) && mySSI.ChartCBEMA) 
            {
                return;
            }

            for (i = 0; i< ChartDisplay.Series.Count; i++)
            {
                EventData.InputSettings.ActiveChannels[i] = ChartDisplay.Series[i].Active;
            }
        }
/************************************************************************
* NAME:    InitChartActiveEventsDefault
* FUNCTIONAL DESCRIPTION:
* Set the default state of active events
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************/
        public override void InitChartActiveEventsDefault()
        {
            int i;

            EventData.InputSettings.SetActiveEventLength(EventData.IndexTable.Count);
            for (i = 0; i<EventData.IndexTable.Count; i++) 
            {
                EventData.InputSettings.ActiveEvents[i] = false;
            }
        }
/************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Reset chart axis limits to leave small interval around limits of data
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************/
        public override void AutoScaleAxes()
        {
            double xMin, xMax, yMin, yMax;
            TTimeSeriesInputSettingsSSI mySSI;

            /*  TeeChart autoscaling causes an exact fit to the data. Since this sometimes not
            *  visually appealing, this scaling function subtracts 5% from the limits of
            *  the horizontal and vertical axes so that chart is clearly visible in the chart
            *  area.*/

            base.AutoScaleAxes(out xMin,out xMax,out yMin,out yMax);

            /*CBEMA chart requires use of a smaller range in the Y direction because the limit
            curves extend far beyond most data in the positive direction (0 to 300).  This
            has the effect of zooming in a small amount in order to see the sag and swell
            regions more clearly.
            */
            mySSI = (TTimeSeriesInputSettingsSSI)EventData.InputSettings;
            if(mySSI.ChartCBEMA) 
            {
                ChartDisplay.Axes.Bottom.SetMinMax(xMin - (xMax - xMin)/20.0, xMax + (xMax - xMin)/100.0);
                ChartDisplay.Axes.Left.SetMinMax(yMin - (yMax - yMin)/20,yMax - (yMax - yMin)/4.0);
            }
            else
            {
                ChartDisplay.Axes.Bottom.SetMinMax(xMin - (xMax - xMin)/20.0, xMax + (xMax - xMin)/20.0);
                ChartDisplay.Axes.Left.SetMinMax(yMin - (yMax - yMin)/20.0,yMax + (yMax - yMin)/20.0);
            }
        }
/************************************************************************
* NAME:   FillEventIndexTable
* FUNCTIONAL DESCRIPTION:  Scan raw data, identify events, create index of events
* INPUT PARAMETERS:
*                    filter: boolean  use filter controls true/false
* OUTPUT PARAMETERS: none
* RETURN:  result : boolean  true if event index constructed, otherwise false
************************************************************************/
        public override Boolean FillEventIndexTable(Boolean filter) 
        {
            DateTime userStartDate, userEndDate, recordDate;
            Boolean more,found_start,found_end,foundP;
            Boolean eventBeginChecked, eventEndChecked;
            int start,stop,estart,estop,oldStop;
            Boolean hasPostDisturbance;
            Boolean stopIncomplete, startIncomplete, oldStartIncomplete, saveEvent;
            int i,j, postCount;
            THMIEventPhase myPhase;
            TTimeSeriesEventType myEventType;
            TTimeSeriesRecord dataRecord;
            double vMax, vMin;
            double[] voltMin = new double[3];
            double[] voltMax = new double[3];
            Boolean foundMinMax;
            GlobalString gs = new GlobalString();
            string recordStatus = PRE_DISTURBANCE;

            /*The raw data is scanned looking for events.  Events are intervals
            within the data set that have special start and end conditions.
            When an event is found, it is added to a list that contains various
            attributes of each event (start date/time, duration, min/max values, etc.)
            */
            EventData.IndexTable.Clear();   // erase old index
            if(EventData.RawDataTable.Count == 0) 
            {
                return false; // no data
            }

            more = true;
            start = 0;
            estart = 0;

            /* "E"nd records encountered : used to test for event after pre- and
            before post-disturbance*/

            hasPostDisturbance = false;
            stopIncomplete = false;
            startIncomplete = false;

            GetFilterSettingsSSI(filter, out eventBeginChecked, out eventEndChecked,
                   out userStartDate, out userEndDate, out myEventType, out myPhase);

            // scan raw data records searching for events that match user filters

            while (more)
            {
                // Look for beginning of next event
                found_start = TestEventStart(filter,eventBeginChecked,eventEndChecked,
                     userStartDate,userEndDate,ref start, ref more);

                // Test for type of event

                if (found_start)   // found a candidate event in user"s range
                {
                    foundMinMax = false;
                    foundP = true;  // start looking for a non-P record

                    if(start >= (EventData.RawDataTable.Count-1))      // no events found
                    {
                        break;
                    }

                    found_end = false;
                    vMax = 0.0;
                    vMin = 1.0e10;

                    for (j = 0; j< voltMin.Length; j++)
                    {
                        voltMax[j] = 0.0;
                        voltMin[j]= 1.0e10;
                    }

                    //     dataRecord = null;  // initialize for compiler
                    estart = start;  // reset start of event

                    for (i = start; i<EventData.RawDataTable.Count; i++)
                    {
                        dataRecord = EventData.RawDataTable[i];   // get indexed record from list
                        recordStatus = dataRecord.Status;
                        recordDate = dataRecord.DateTime;
                        if(filter && eventEndChecked && (gs.CompareDateTime(recordDate,userEndDate) > 0)) break;  // past end of interval

                        // Look for end of this event

                        found_end = TestEventEnd(i,dataRecord,ref foundP,ref estart,
                                       ref hasPostDisturbance, ref stopIncomplete);
                        if(found_end) 
                        {
                            break;
                        }

                        if(!foundP && (hasPostDisturbance)) 
                        {
                            foundMinMax = true;   // update the min/max values
                            UpdateVoltMinMax(dataRecord,ref voltMin,ref voltMax);
                        }
                    }

                    // Assign the appropriate min/max value to the record
                    FindVoltMinMax(myPhase,foundMinMax,ref voltMin,ref voltMax,ref vMin,ref vMax);

                    estop = i;   // end of event - {ning of post-distrubance phase

                    postCount = 0;
                    if((i < EventData.RawDataTable.Count) &&
                        (recordStatus != PRE_DISTURBANCE))    // scan to end of post-disturbance
                    {
                        for (j = i; j<EventData.RawDataTable.Count; j++)
                        {
                            dataRecord = EventData.RawDataTable[j];
                            if (dataRecord.Status != POST_DISTURBANCE)  break;  // retain post-disturbance in this event
                            postCount++;  // count actual post-disturbance records
                        }
                        i = j;  // end of post-disturbance
                    }

                    if(hasPostDisturbance)   // test for special interrupt condition
                    {
                        stop = i-1;   // use previous record as the last record in this event
                    }
                    else
                    {
                        stop = i;     // Use this record as the last record in this event
                    }

                    oldStartIncomplete = startIncomplete;   // incoming state for this event
                    oldStop = stop;  // end of this event, if incomplete event, stop is advanced
                    saveEvent = TestEventIncomplete(ref startIncomplete, ref stopIncomplete, ref stop);

                    // found a useful event or incomplete event at end of dataset, record it in the index
                    if(saveEvent &&
                        ((found_end || (stop >= (EventData.RawDataTable.Count-1))))) 
                    {
                        CreateNewEventRecord(found_end,estart,estop,start,oldStop,
                                     voltMin,voltMax,vMin,vMax,
                                     oldStartIncomplete,stopIncomplete);
                    }

                    start = stop;  // new record to start search from
                    if(hasPostDisturbance && (postCount == 0))
                    {
                        start++;  // start new event from following record
                    }
                    if (!found_end || (start >= (EventData.RawDataTable.Count-1))) 
                    {
                        more = false;  // no more records to examine
                    }
                }
                else
                {
                    more = false;  // no more records to examine
                }
            }

            MergeEvents();  // merge sag events that were interrupted
            FilterEvents(myEventType, myPhase); // Delete filtered events after merging sub-events

            return true; // if we got here, it is ok to generate event list and chart
        }
/************************************************************************
* NAME:  GetFilterSettingsSSI
* FUNCTIONAL DESCRIPTION:  Get the date/time, phase, event filter settings
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS:
*                    event{Checked : boolean   start date/time specified
*                    eventEndChecked : boolean     end date/time specified
*                    userStartDate   : TDateTime   start date/time
*                    userEndDate     : TDateTime   end date/time
*                    myEventType : TTimeSeriesEventType   type of event to find
*                    myPhase : THMIEventPhase   phase to identify event on
* RETURN: none
************************************************************************/
        public void GetFilterSettingsSSI(Boolean filter,
                       out Boolean eventBeginChecked, out Boolean eventEndChecked,
                       out DateTime userStartDate, out DateTime userEndDate,
                       out TTimeSeriesEventType myEventType,
                       out THMIEventPhase myPhase)
        {
            TTimeSeriesInputSettingsSSI myInputSettings;

            myInputSettings = (EventData.InputSettings as TTimeSeriesInputSettingsSSI);

            // get users supplied start/end dates in case they are active
            userStartDate = myInputSettings.BeginDateTime;
            userEndDate = myInputSettings.EndDateTime;
            eventBeginChecked = myInputSettings.BeginDateActive;
            eventEndChecked = myInputSettings.EndDateActive;
            // phase and event types selected by user
            if(filter)     // filters are active
            {
                myPhase = myInputSettings.EventPhase;
                myEventType = myInputSettings.EventType;
            }
            else
            {              // ignore filters
                myPhase = THMIEventPhase.epAll; // all
                myEventType = TTimeSeriesEventType.evAll; // all
            }
        }
/************************************************************************
* NAME:  TestEventStart
* FUNCTIONAL DESCRIPTION:  Identify start of next event
* INPUT PARAMETERS:  filter : boolean   use date/time filters
*                    event{Checked : boolean  test start date/time
*                    eventEndChecked : boolean    test end date/time
*                    userStartDate : TDateTime    start date/time
*                    userEndDate   : TDateTime    end date/time
* OUTPUT PARAMETERS: start : integer   starting record number
*                    more : boolean   if true continue scan, otherwise stop
* RETURN: result : boolean   true if start of event found
************************************************************************/
        public Boolean TestEventStart(Boolean filter, Boolean eventBeginChecked,
                     Boolean eventEndChecked ,
                     DateTime userStartDate, DateTime userEndDate,
                     ref int start, ref Boolean more)
        {
            Boolean found_start;
            int i, j;
            TTimeSeriesRecord dataRecord, dataRecord1;
            DateTime recordDate;
            TTimeSeriesEventType eventType;
            GlobalString gs = new GlobalString();

            found_start = false;  // flag set if new event found

            // look for next event starting where last event ended, to end of dataset
            for (i = start; i<EventData.RawDataTable.Count; i++)
            {
                dataRecord = EventData.RawDataTable[i];
                if(dataRecord.Status == INVALID_RECORD) continue;
                recordDate = dataRecord.DateTime;

                if (filter && eventBeginChecked && (gs.CompareDateTime(recordDate,userStartDate) < 0)) 
                {
                    continue;  // before { of interval
                }
                if (filter && eventEndChecked && (gs.CompareDateTime(recordDate,userEndDate) > 0)) 
                {
                    more = false;  // no more useful data to examine
                    break;  // past end of interval
                }

                if(dataRecord.Status == POST_DISTURBANCE) continue;

                // test for immediate transition from sag to swell or swell to sag
                // if i is first in event but not first in dataset

                if((i == start) && (i > 0) && (dataRecord.Phase != NOMINAL_3P)) 
                {
                    eventType = PhaseToEventType(i-1);   // test type of previous record
                    switch (eventType)
                    {
                        case TTimeSeriesEventType.evSwell:          // going from swell to sag?
                            if(!gs.ContainsText(dataRecord.Phase,SWELL_EVENT))    // swell ended
                            {
                                found_start = true;
                            }
                        break;

                        case TTimeSeriesEventType.evSag:            // going from sag to swell?
                            if(!(gs.ContainsText(dataRecord.Phase,SAG_EVENT) ||
                                gs.ContainsText(dataRecord.Phase,INTERRUPT_EVENT)))   // sag ended
                            {
                               found_start = true;
                            }
                        break;
                    }
                }

                if(!found_start) 
                {
                    for (j = PHASE_A; j<= PHASE_C; j++)
                    {
                        if(dataRecord.Phase[j] != NOMINAL_EVENT) 
                        {
                            found_start = true;
                            break;
                        }
                    }
                }

                if(!found_start) 
                {
                    // test for Trigger event
                    if((dataRecord.Phase == NOMINAL_3P) && (dataRecord.Status == FAST_RECORD)) 
                    {
                        found_start = true;
                    }
                    else
                    {    // test for event following a trigger
                        found_start = TestEventPreviousTrigger(i,start,dataRecord);
                        if(!found_start && (i > 0))   // test previous for interrupt
                        {
                            dataRecord1 = EventData.RawDataTable[i-1];
                            if((dataRecord1.Phase == INTERRUPT_3P) && (dataRecord.Phase != INTERRUPT_3P)) 
                            {
                                found_start = true;
                            }
                        }
                    }

                    if(!found_start) 
                    {
                        if(dataRecord.Phase == INTERRUPT_3P) 
                        {
                            found_start = true;  // 3-phase interrupt
                        }
                        else if (dataRecord.Status == PRE_DISTURBANCE)    // test for pre-disturbance record
                        {
                            found_start = true;
                        }
                    }
                }

                if(found_start) 
                {
                    start = i;    // index of candidate event
                    break;
                }
            }
            return found_start;
       }
/************************************************************************
* NAME:  TestEventEnd
* FUNCTIONAL DESCRIPTION:  Identify end of present event
* INPUT PARAMETERS:  iRec : integer    record number
*                    dataRecord : TTimeSeriesRecord   SSI record
* OUTPUT PARAMETERS: foundP : boolean   a P record found
*                    estart : integer   first record of event
*                    hasPostDisturbance : boolean  true if event has post-dist. recs.
*                    stopIncomplete : boolean true if event end is incomplete
* RETURN: result : boolean   true if end of event found
************************************************************************/
        public Boolean TestEventEnd(int iRec,
                   TTimeSeriesRecord dataRecord,
                   ref Boolean foundP,
                   ref int eStart,
                   ref Boolean hasPostDisturbance, ref Boolean stopIncomplete)
        {
            Boolean found_end;
            TTimeSeriesRecord dataRecordPrev;
            TTimeSeriesEventType eventType;
            Boolean foundPostDisturbance;
            GlobalString gs = new GlobalString();

            found_end = false;
            hasPostDisturbance = true;  // assume event will end with post-disturbance records
            foundPostDisturbance = false;  // assume no post-disturbance found
            stopIncomplete = false;   // assume the event is complete

            // test for end of pre-disturbance
            if(foundP && (dataRecord.Status != PRE_DISTURBANCE) && 
                (dataRecord.Status != READY_RECORD)) // R=Meter powered up
            {
                foundP = false;
                eStart = iRec;  // end of pre-disturbance (start of actual event)
            }

            // test for end of event

            if(dataRecord.Status == INVALID_RECORD)    // invalid (overflow) record
            {
                found_end = true;
                hasPostDisturbance = false;
                stopIncomplete = true;
            }
            else
            {
                if((iRec > 0) && (eStart != iRec))      // test special event termination conditions
                {
                    dataRecordPrev = EventData.RawDataTable[iRec-1];
                    if((iRec > eStart) && (dataRecordPrev.Phase == INTERRUPT_3P) &&
                            (dataRecord.Phase != INTERRUPT_3P)) 
                    {
                        found_end = true;  // 3-phase interrupt is over
                        hasPostDisturbance = false;      // no post-disturbance
                    }
                    else if((iRec > eStart) && (dataRecordPrev.Phase != INTERRUPT_3P) &&
                            (dataRecord.Phase == INTERRUPT_3P)) 
                    {
                        found_end = true;  // 3-phase interrupt started
                        hasPostDisturbance = false;      // no post-disturbance
                    }
                    else if((dataRecordPrev.Phase != NOMINAL_3P) && (dataRecord.Phase == NOMINAL_3P) &&
                            (dataRecord.Status != PRE_DISTURBANCE) && (dataRecord.Status != READY_RECORD)) 
                    {
                        hasPostDisturbance = true;   // found end of event with post-disturbance
                    }
                    else if (dataRecord.DateTime < dataRecordPrev.DateTime)
                    {
                        found_end = true;  // 3-phase interrupt started
                        hasPostDisturbance = true;      // end event with previous record
                        stopIncomplete = true;
                    }
                }

                if(dataRecord.Status == POST_DISTURBANCE) 
                {
                    found_end = true;
                    foundPostDisturbance = true;
                    hasPostDisturbance = true;   // found end of event with post-disturbance
                }

                // test for direct transition from sag to swell or swell to sag

                if(!found_end && (dataRecord.Phase != NOMINAL_3P) && (iRec > eStart)) 
                {
                    eventType = PhaseToEventType(eStart);
                    switch (eventType)
                    {
                        case TTimeSeriesEventType.evSwell:     // transition from swell to sag

                        if(!gs.ContainsText(dataRecord.Phase,SWELL_EVENT))    // swell ended
                        {
                            found_end = true;
                            hasPostDisturbance = false;      // no post-disturbance
                        }
                        break;

                        case TTimeSeriesEventType.evSag:       // transition from sag to swell

                        if(!(gs.ContainsText(dataRecord.Phase,SAG_EVENT) ||
                           gs.ContainsText(dataRecord.Phase,INTERRUPT_EVENT)))   // sag ended
                        {
                            found_end = true;
                            hasPostDisturbance = false;      // no post-disturbance
                        }
                        break;
                    }
                }
            }

            // test for standard event termination

            if(!foundP) 
            {
                Boolean nominal;
                if ((iRec > 0) && (eStart != iRec))
                {
                    dataRecordPrev = EventData.RawDataTable[iRec - 1];
                    nominal = (dataRecordPrev.Phase != NOMINAL_3P) && (dataRecord.Phase == NOMINAL_3P);
                }
                else
                {
                    nominal = false;
                }
                if((dataRecord.Status == PRE_DISTURBANCE) || (dataRecord.Status == READY_RECORD) || nominal) 
                {
                    found_end = true;    // end of event
                    hasPostDisturbance = foundPostDisturbance;
                }
            }
            return found_end;
        }
/************************************************************************
* NAME:  TestEventIncomplete
* FUNCTIONAL DESCRIPTION:  Identify end of present event
* INPUT PARAMETERS:  startIncomplete : boolean  true if previous event was incomplete
*                    stopIncomplete : boolean true if next event is incomplete
* OUTPUT PARAMETERS: irec : integer   record number at end of an event
* RETURN: result : boolean   true if this event should be kept
************************************************************************/
        public Boolean TestEventIncomplete(
                        ref Boolean startIncomplete, ref Boolean stopIncomplete,
                        ref int irec) 
        {
            TTimeSeriesRecord dataRecord;
            Boolean saveEvent;
            string oldStatus;
            DateTime dateTime1;

            saveEvent = true;   // assume this event will be saved
            startIncomplete = stopIncomplete;   // assume it is the same

            if(stopIncomplete)     // previous event was incomplete. Test next event
            {
                dataRecord = EventData.RawDataTable[irec];
                oldStatus = dataRecord.Status;  // remember starting status
                dateTime1 = dataRecord.DateTime;  // remember starting date/time
                irec++;
                dataRecord = EventData.RawDataTable[irec];
                if (dataRecord.DateTime < dateTime1)  // Date/Time setback, terminate event
                {
                    irec--;  // retain original termination
                    return saveEvent;
                }
                while((dataRecord.Status == INVALID_RECORD) &&
                        (irec < (EventData.RawDataTable.Count-1))) 
                {
                    irec++;
                    dataRecord = EventData.RawDataTable[irec];
                }
                if(dataRecord.Phase == NOMINAL_3P)   // next event might be complete
                {
                    startIncomplete = false;  // next event is not incomplete
                    if((oldStatus == PRE_DISTURBANCE) || (oldStatus == READY_RECORD)) 
                    {
                        saveEvent = false;  // don"t the keep leading pre-disturbance
                        stopIncomplete = false;  // just ignore this event
                        startIncomplete = false;
                        if(dataRecord.Status == POST_DISTURBANCE) 
                        {
                            saveEvent = false;   // advance to end of post-disturbance
                            while((dataRecord.Status == POST_DISTURBANCE) &&
                                (irec < (EventData.RawDataTable.Count-1))) 
                            {
                                irec++;
                                dataRecord = EventData.RawDataTable[irec];
                            }
                        }
                    }
                    else      // next event is POST_DISTURBANCE
                    {         // NOTE: THIS CANNOT HAPPEN (CODE IS FOR COMPLETENESS)
                        saveEvent = true;     // keep the leading post-disturbance event
                        if(dataRecord.Status == POST_DISTURBANCE)    // trailing event
                        {
                            while((dataRecord.Status == POST_DISTURBANCE) &&  // advance to end of post-disturbance
                                (irec < (EventData.RawDataTable.Count-1))) 
                            {
                                irec++;
                                dataRecord = EventData.RawDataTable[irec];
                            }
                        }
                    }
                }
            }
            return saveEvent;
        }
/************************************************************************
* NAME:  TestEventPreviousTrigger
* FUNCTIONAL DESCRIPTION:  Test transition from previous event
* INPUT PARAMETERS:  iRec : integer    record number
*                    start : first record of event
*                    dataRecord : TTimeSeriesRecord   SSI record
* OUTPUT PARAMETERS: none
* RETURN: result : boolean   true if start of event found
************************************************************************/
        public Boolean TestEventPreviousTrigger(int iRec, int start,
                            TTimeSeriesRecord dataRecord)
        {
            TTimeSeriesRecord dataRecordPrevious;
            Boolean found_start;
            int j;

            found_start = false;
            if(iRec > start) 
            {
                // examine previous record
                dataRecordPrevious = EventData.RawDataTable[iRec-1];
                if(dataRecordPrevious.Phase == NOMINAL_3P)    // previous was trigger record
                {
                    if(dataRecord.Phase == INTERRUPT_3P)   // Interrupt detected
                    {
                        found_start = true;
                    }
                    else
                    {
                        for (j = PHASE_A; j<= PHASE_C ; j++) // transition to a sag or swell on any phase
                        {
                            if((dataRecord.Phase[j] == SWELL_EVENT) ||
                                (dataRecord.Phase[j] == SAG_EVENT) ||
                                (dataRecord.Phase[j] == INTERRUPT_EVENT)) 
                            {
                                found_start = true;
                                break;
                            }
                        }
                    }
                }
            }
            return found_start;
        }
/************************************************************************
* NAME:  FindSinglePhaseEvent
* FUNCTIONAL DESCRIPTION:  Find a single phase event
* INPUT PARAMETERS: myEventType : TTimeSeriesEventType   type of event to find
*                   myPhase : THMIEventPhase   phase to identify event on
*                   dataRecord : TTimeSeriesRecord   SSI record
* OUTPUT PARAMETERS: foundO : boolean   found a swell event
*                    foundU : boolean   found a sag event
*                    foundI : boolean   found an interrupt event
*                    foundA : boolean   found 3-phase event
* RETURN: none
************************************************************************/
        public void FindSinglePhaseEvent(TTimeSeriesEventType myEventType,
               THMIEventPhase myPhase,
               TTimeSeriesRecord dataRecord,
               ref Boolean foundO,ref Boolean foundU,ref Boolean foundI,ref Boolean foundA)
        {
            switch (myEventType) 
            {
                case TTimeSeriesEventType.evInterrupt:
                    if (dataRecord.Phase[(int)myPhase+1] == INTERRUPT_EVENT) foundI = true;
                break;

                case TTimeSeriesEventType.evSag:  // sag
                    if (dataRecord.Phase[(int)myPhase+1] == SAG_EVENT) foundU = true;
                    // an Interrupt is a severe sag
                    if (dataRecord.Phase[(int)myPhase+1] == INTERRUPT_EVENT) foundI = true;
                break;

                case TTimeSeriesEventType.evSwell:  // swell
                    if (dataRecord.Phase[(int)myPhase+1] == SWELL_EVENT) foundO = true;
                break;

                case TTimeSeriesEventType.evAll:  // All
                    if (dataRecord.Phase[(int)myPhase+1] != NOMINAL_EVENT) foundA = true;
                break;

                default:
                    System.Diagnostics.Debug.Assert(false, "unknown event type");
                break;
            }
        }
/************************************************************************
* NAME:  FindAnyPhaseEvent
* FUNCTIONAL DESCRIPTION:  Find any event (no filtering for event type)
* INPUT PARAMETERS: myEventType : TTimeSeriesEventType
*                   dataRecord : TTimeSeriesRecord   SSI record
* OUTPUT PARAMETERS: foundO : boolean   found a swell event
*                    foundU : boolean   found a sag event
*                    foundI : boolean   found an interrupt event
*                    foundA : boolean   found 3-phase event
* RETURN: none
************************************************************************/
        public void FindAnyPhaseEvent(TTimeSeriesEventType myEventType,
                   TTimeSeriesRecord dataRecord,
                   ref Boolean foundO,ref Boolean foundU,ref Boolean foundI,ref Boolean foundA)
        {
            switch (myEventType)
            {
                case TTimeSeriesEventType.evInterrupt: // Interrupt
                    if ((dataRecord.Phase[PHASE_A] == INTERRUPT_EVENT) ||
                        (dataRecord.Phase[PHASE_B] == INTERRUPT_EVENT) ||
                        (dataRecord.Phase[PHASE_C] == INTERRUPT_EVENT)) foundI = true;
                break;

                case TTimeSeriesEventType.evSag:    // Sag
                    if ((dataRecord.Phase[PHASE_A] == SAG_EVENT) ||
                        (dataRecord.Phase[PHASE_B] == SAG_EVENT) ||
                        (dataRecord.Phase[PHASE_C] == SAG_EVENT)) foundU = true;
                    // an Interrupt is a severe sag
                    if ((dataRecord.Phase[PHASE_A] == INTERRUPT_EVENT) ||
                        (dataRecord.Phase[PHASE_B] == INTERRUPT_EVENT) ||
                        (dataRecord.Phase[PHASE_C] == INTERRUPT_EVENT)) foundI = true;
                break;

                case TTimeSeriesEventType.evSwell: // swell
                    if ((dataRecord.Phase[PHASE_A] == SWELL_EVENT) ||
                        (dataRecord.Phase[PHASE_B] == SWELL_EVENT) ||
                        (dataRecord.Phase[PHASE_C] == SWELL_EVENT)) foundO = true;
                break;

                case TTimeSeriesEventType.evAll: // All
                    if ((dataRecord.Phase[PHASE_A] != NOMINAL_EVENT) ||
                        (dataRecord.Phase[PHASE_B] != NOMINAL_EVENT) ||
                        (dataRecord.Phase[PHASE_C] != NOMINAL_EVENT)) foundA = true;
                    if (foundA && ((dataRecord.Phase[1] != SWELL_EVENT) ||
                        (dataRecord.Phase[PHASE_B] != SWELL_EVENT) ||
                        (dataRecord.Phase[PHASE_C] != SWELL_EVENT))) foundU = true; // trigger min calculation
                    if ((dataRecord.Phase[PHASE_A] == INTERRUPT_EVENT) ||
                        (dataRecord.Phase[PHASE_B] == INTERRUPT_EVENT) ||
                        (dataRecord.Phase[PHASE_C] == INTERRUPT_EVENT)) foundI = true;
                break;
                default:
                    System.Diagnostics.Debug.Assert(false, "unknown event type");
                break;
            }
        }
/************************************************************************
* NAME:  UpdateVoltMinMax
* FUNCTIONAL DESCRIPTION:  Update voltage min/max from data record
* INPUT PARAMETERS:  dataRecord : TTimeSeriesRecord  data record
* OUTPUT PARAMETERS:  voltMin : array of double   new minimum array for each phase
*                     voltMax : array of double   new maximum array for each phase
* RETURN: none
************************************************************************/
        public void UpdateVoltMinMax(TTimeSeriesRecord dataRecord,
                   ref double[] voltMin,ref double[] voltMax)
        {
            const int FIRST_VOLTAGE_4XX = 0;
            const int FIRST_VOLTAGE = 5;

            int j;
            int voltOffset;

            System.Diagnostics.Debug.Assert((voltMin.Length == voltMax.Length), "Inconsistent array lengths");
            if(EventData.RawDataLabels.Count < 7)      // SEL4xx
            {
                voltOffset = FIRST_VOLTAGE_4XX;
            }
            else
            {
                voltOffset = FIRST_VOLTAGE;
            }
            // record min/max for each voltage
            for( j =0; j< voltMin.Length; j++)
            {
                if(dataRecord.Value[j + voltOffset] < voltMin[j]) 
                {
                    voltMin[j] = dataRecord.Value[j + voltOffset];
                }
                if(dataRecord.Value[j + voltOffset] > voltMax[j]) 
                {
                    voltMax[j] = dataRecord.Value[j + voltOffset];
                }
            }
        }
/************************************************************************
* NAME:  FindVoltMinMax
* FUNCTIONAL DESCRIPTION:  Find voltage min/max
* INPUT PARAMETERS:
*                   myPhase : THMIEventPhase   phase to identify event on
*                   foundMinMax : boolean  if true, update with new min/max values
* OUTPUT PARAMETERS: voltMin : double  minimum voltage on each phase
*                    voltMax : double  maximum voltage on each phase
*                    vMin : double   minimum voltage all phases
*                    vMax : double   maximum voltage all phases
* RETURN: none
************************************************************************/
        public void FindVoltMinMax(THMIEventPhase myPhase,
                   Boolean foundMinMax,
                   ref double[] voltMin, ref double[] voltMax,
                   ref double vMin, ref double vMax)
        {
            int j;

            System.Diagnostics.Debug.Assert((voltMin.Length == voltMax.Length), "Invalid array lengths.");

            if(foundMinMax) 
            {
                switch (myPhase)                // assign filtered min/max
                {
                    case THMIEventPhase.epA:
                    case THMIEventPhase.epB:
                    case THMIEventPhase.epC:  //a,b,c
                        vMin = voltMin[(int)myPhase];
                        vMax = voltMax[(int)myPhase];
                    break;

                    case THMIEventPhase. epAll:      // all
                        for( j = 0; j< voltMin.Length; j++) 
                        {
                            if(vMin > voltMin[j]) 
                            {
                                vMin = voltMin[j];
                            }
                            if(vMax < voltMax[j]) 
                            {
                                vMax = voltMax[j];
                            }
                        }
                    break;

                    default:
                        System.Diagnostics.Debug.Assert(false,"Invalid phase filter.");
                    break;
                }
            }
            else
            {
                vMin = 100.0;
                vMax = 100.0;
                for (j = 0; j< voltMin.Length; j++)
                {
                    voltMin[j] = 100.0;
                    voltMax[j] = 100.0;
                }
            }
        }
/************************************************************************
* NAME:  ResetUnwantedEvents
* FUNCTIONAL DESCRIPTION:  Cancel event flags for unwanted events
* INPUT PARAMETERS: myEventType : TTimeSeriesEventType
* OUTPUT PARAMETERS:
*                   foundO : boolean   flag indicated swell event found
*                   foundU : boolean   flag indicated sag event found
*                   foundI : boolean   flag indicated interrupt event found
* RETURN: none
************************************************************************/
        public void ResetUnwantedEvents(TTimeSeriesEventType myEventType,
                  ref Boolean foundO, ref Boolean foundU, ref Boolean foundI)
        {
            switch (myEventType)
            {
                case TTimeSeriesEventType.evInterrupt: // Interrupt
                    foundU = false;     // cancel sag and swell events
                    foundO = false;
                break;

                case TTimeSeriesEventType.evSag:    // Sag
                    foundI = false;    // cancel interrupt and swell events
                    foundO = false;
                break;

                case TTimeSeriesEventType.evSwell: // swell
                    foundI = false;    // cancel interrupt and sag events
                    foundU = false;
                break;

                case TTimeSeriesEventType.evAll: // All
                    // do nothing (user wants all events)
                break;

                default:
                    System.Diagnostics.Debug.Assert(false, "unknown event type");
                break;
            }
        }
/************************************************************************
* NAME:  CreateNewEventRecord
* FUNCTIONAL DESCRIPTION:  Create a new record in the event index
* INPUT PARAMETERS: found_end : boolean  found the end of an event
*                   estart : integer   start of event after pre-disturbance
*                   estop  : integer   end of event before post-disturbance
*                   start  : integer   start of event pre-disturbance
*                   stop   : integer   end of event post-disturbance
*                   voltMin : array of double  minimum voltage each phase
*                   voltMax : array of double  maximum voltage each phase
*                   vMin : double  minimum voltage all phases
*                   vMax : double  maximum voltage all phases
*                   isIncomplete : boolean  true if event is incomplete
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void CreateNewEventRecord(Boolean found_end,
                    int estart,int estop,int start,int stop,
                    double[] voltMin, double[] voltMax,
                    double vMin, double vMax,
                    Boolean startIncomplete, Boolean stopIncomplete)
        {
            const string EVENT_ERROR = "Error during event record creation";

            DateTime dt1, dt2;
            TTimeSeriesIndexRecord indexRecord;
            TTimeSeriesRecord dataRecord;

            indexRecord = new TTimeSeriesIndexRecord();  // new index record
            try
            {
                dataRecord = EventData.RawDataTable[estart]; // first record of event
                dt1 = dataRecord.DateTime;
                indexRecord.DateTime = dataRecord.DateTime;
                if(stop < estop) 
                {
                    estop = stop;
                }
                indexRecord.FirstRecord = start;
                indexRecord.LastRecord = stop;
                indexRecord.StartEvent = estart;
                indexRecord.StopEvent = estop;
                indexRecord.EventName = "#" + string.Format("{0:d4}",EventData.IndexTable.Count+1);
                if(startIncomplete) 
                {
                    if(stopIncomplete) 
                    {
                        indexRecord.Completion = TTimeSeriesCompleteType.ctIncompleteBoth;
                    }
                    else
                    {
                        indexRecord.Completion = TTimeSeriesCompleteType.ctIncompleteStart;
                    }
                }
                else if(stopIncomplete) 
                {
                    indexRecord.Completion = TTimeSeriesCompleteType.ctIncompleteStop;
                }
                else
                {
                    indexRecord.Completion = TTimeSeriesCompleteType.ctComplete;
                }

                dataRecord = EventData.RawDataTable[estop]; // last record of event
                dt2 = dataRecord.DateTime;

                indexRecord.Duration = dt2.Subtract(dt1);
                indexRecord.Checked = false;  // set default event flag as not active in display
                indexRecord.EventType = PhaseToEventType(estart);
                indexRecord.Description = PhaseToEventDescription(indexRecord.EventType);
                if(!found_end || startIncomplete || stopIncomplete) 
                {
                    indexRecord.Description = indexRecord.Description + "**";
                    indexRecord.Completion = TTimeSeriesCompleteType.ctIncompleteStop;
                }

                indexRecord.VAMin = voltMin[0];
                indexRecord.VAMax = voltMax[0];
                indexRecord.VBMin = voltMin[1];
                indexRecord.VBMax = voltMax[1];
                indexRecord.VCMin = voltMin[2];
                indexRecord.VCMax = voltMax[2];

                switch (indexRecord.EventType)
                {
                    case TTimeSeriesEventType.evNone:
                    case TTimeSeriesEventType.evSag:
                    case TTimeSeriesEventType.evMulti:
                    case TTimeSeriesEventType.evInterrupt:
                        indexRecord.Magnitude = vMin;
                    break;

                    case TTimeSeriesEventType.evSwell:
                        indexRecord.Magnitude = vMax;
                    break;

                    default:
                        System.Diagnostics.Debug.Assert(false,"Invalid violation type");
                    break;
                }

                // get CBEMA violation code if any

                indexRecord.Violation = CBEMAViolation(indexRecord);
                EventData.IndexTable.Add(indexRecord);  // add new record into event index
            }
            catch
            {
                EventData.IndexTable.Remove(EventData.IndexTable[EventData.IndexTable.Count-1]); // delete bad record
                throw new Exception(EVENT_ERROR);
            }
        }
/************************************************************************
* NAME:  MergeEvents
* FUNCTIONAL DESCRIPTION:  Merge interrupted sag events
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN: result : none
************************************************************************/
        public void MergeEvents()
        {
            TTimeSeriesIndexRecord indexRecord,indexRecord2;
            TTimeSeriesRecord dataRecord;
            TimeSpan duration;
            int i,j,k,first,last,count,estop,irec,irec2;
            int[] deleteList;  // list of events to delete
            Boolean foundSag, foundInterrupt;

            deleteList = new int[EventData.IndexTable.Count];  // worst case length

            i = 0;
            count = 0;
            first = 0;         // first event in the sequence
            last = 0;          // last event in the sequence
            duration = new TimeSpan(0);    // total duration of the merged sag event
            foundInterrupt = false;

            /* Look for sequences of events that start with a sag followed by one or
            more interrupts. If the terminating event is a sag, then eliminate the
            final sag and reassign the initial sag attributes to span the complete
            interval.  If the terminating event is neither a sag nor an interrupt, then
            merge only the interrupt events into the initial sag.

            If an overwrite occurs during an interrupt event, then
            two parent events result: one that starts with a sag if any and ends with the
            leading interrupt, and one that {s with the trailing interrupt
            event and ends with the trailing sag event if any. If the trailing interrupt
            ends with a sag, then the sag must be updated to include the interrupt as
            a sub-event.
            */

            while (i < EventData.IndexTable.Count-1)
            {
                indexRecord = EventData.IndexTable[i]; // first sag event
                indexRecord2 = EventData.IndexTable[i]; // first sag event
                irec = i;
                irec2 = i;
                foundSag = false;
                if((indexRecord.EventType == TTimeSeriesEventType.evSag) ||
                    ((indexRecord.EventType == TTimeSeriesEventType.evInterrupt) &&
                    (indexRecord.Completion == TTimeSeriesCompleteType.ctIncompleteStart)))   // candidate for merge
                {
                    first = i;    // start of sag event
                    last = first;
                    j = first + 1;
                    duration = indexRecord.Duration;  // start accumulating duration
                    foundInterrupt = (indexRecord.EventType == TTimeSeriesEventType.evInterrupt);
                    while (j < EventData.IndexTable.Count )
                    {
                        indexRecord2 = EventData.IndexTable[j]; // next interrupt
                        irec2 = j;
                        last = j;  // remember last event
                        if(indexRecord2.Completion != TTimeSeriesCompleteType.ctComplete)    // flag parent event
                        {
                            indexRecord.Completion = indexRecord2.Completion;
                            indexRecord.Description = indexRecord.Description + "**";
                        }

                        // see if trailing sag will span preceding interrupt
                        if((indexRecord2.EventType == TTimeSeriesEventType.evSag) &&
                            ((last - first) == 1) && // second event
                            (indexRecord.EventType == TTimeSeriesEventType.evInterrupt)) 
                        {
                            duration = duration + indexRecord2.Duration;
                            indexRecord2.Duration = duration;
                            indexRecord2.FirstRecord = indexRecord.FirstRecord;
                            indexRecord2.StartEvent = indexRecord.StartEvent;
                            indexRecord2.DateTime = indexRecord.DateTime;
                            MergeMinMaxVoltages(ref indexRecord,indexRecord2);
                            EventData.IndexTable[irec2] = indexRecord2;  // replace the updated record
                            // swap events and make this the parent event
                            indexRecord = EventData.IndexTable[j];
                            EventData.IndexTable.Reverse(first, j);  
                            irec = first;
                        }
                        else if(indexRecord2.EventType == TTimeSeriesEventType.evInterrupt) 
                        {
                            foundInterrupt = true;
                            MergeMinMaxVoltages(ref indexRecord,indexRecord2);
                            duration = duration + indexRecord2.Duration;  // increment duration
                            if((indexRecord2.Completion == TTimeSeriesCompleteType.ctIncompleteStop) ||
                                (indexRecord2.Completion == TTimeSeriesCompleteType.ctIncompleteBoth))   // terminate now
                            {
                                last++;  // increment final sub-event 
                                i++;
                                break;
                            }
                        }
                        else if((indexRecord2.EventType == TTimeSeriesEventType.evSag) && foundInterrupt) 
                        {    // sag condition, but maybe not termination of initial sag
                            foundSag = true;  // assume it is a true sag returning to nominal
                            estop = indexRecord2.StopEvent; // last record of sag interval
                            dataRecord = EventData.RawDataTable[estop]; // first record of event
                            for (k = 1; k<= 3 ; k++)    // see if sag reaches nominal state
                            {
                                if((dataRecord.Phase[k] == 'I') || (dataRecord.Phase[k] == 'U')) 
                                {
                                    foundSag = false;  // not nominal yet so not terminal sag
                                    duration = duration + indexRecord2.Duration;  // increment duration
                                    deleteList[count] = j;    // delete this sub-sag
                                    count++;
                                    break;
                                }
                            }
                            if(foundSag ||  // found terminating sag, resume search for next parent event
                                (indexRecord2.Completion != TTimeSeriesCompleteType.ctComplete)) 
                            {
                                i++;
                                break;  // reached end of sag with sub-events
                            }
                        }
                        if(!foundInterrupt || (indexRecord2.EventType == TTimeSeriesEventType.evSwell)) 
                        {
                            break;   // reached end of sag with sub-events;
                        }
                        i++;
                        j++;

                        if((indexRecord2.Completion == TTimeSeriesCompleteType.ctIncompleteStop) ||
                            (indexRecord2.Completion == TTimeSeriesCompleteType.ctIncompleteBoth))   // terminate now
                        {
                            break;
                        }
                    }
                    if(((foundSag && ((last - first) > 1))) ||    // not next event and not interrupt
                        ((indexRecord2.EventType == TTimeSeriesEventType.evInterrupt) &&
                        ((indexRecord2.Completion == TTimeSeriesCompleteType.ctIncompleteStop) ||
                        (indexRecord2.Completion == TTimeSeriesCompleteType.ctIncompleteBoth))))   // terminate now
                    {
                        if(indexRecord2.EventType == TTimeSeriesEventType.evSag)  // merge with first event and delete
                        {
                            indexRecord.Duration = duration + indexRecord2.Duration;
                            indexRecord.LastRecord = indexRecord2.LastRecord;
                            indexRecord.StopEvent = indexRecord2.StopEvent;
                            deleteList[count] = last;
                            count++;
                        }
                        else        // merge first sag with last interrupt event (no deletion)
                        {
                            indexRecord2 = EventData.IndexTable[last-1]; // last interrupt
                            indexRecord.Duration = duration; // duration already correct
                            indexRecord.LastRecord = indexRecord2.LastRecord;
                            indexRecord.StopEvent = indexRecord2.StopEvent;
                        }
                        MergeMinMaxVoltages(ref indexRecord,indexRecord2);
                        EventData.IndexTable[irec] = indexRecord;  // replace the updated record
                        first = 0;
                        last = 0;   // reset flag to facilitate test later
                    }
                }
                i++;
            }

            if(foundInterrupt && (last > first))   // there was a sag followed by interrupt but no terminating event
            {
                indexRecord = EventData.IndexTable[first]; // initial sag event
                indexRecord2 = EventData.IndexTable[last]; // last interrupt event
                indexRecord.Duration = duration;
                indexRecord.LastRecord = indexRecord2.LastRecord;
                indexRecord.StopEvent = indexRecord2.StopEvent;
                MergeMinMaxVoltages(ref indexRecord,indexRecord2);
                EventData.IndexTable[first] = indexRecord;  // replace the updated record
            }
            DeleteEvents(deleteList,count);
        }
/************************************************************************
* NAME: MergeMinMaxVoltages
* FUNCTIONAL DESCRIPTION: Update parent event with sub-event minimum voltages
* INPUT PARAMETERS: indexRecordP : TTimeSeriesIndexRecord   parent event index record
*                   indexRecordS : TTimeSeriesIndexRecord   sub-event index record
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public void MergeMinMaxVoltages(ref TTimeSeriesIndexRecord indexRecordP,TTimeSeriesIndexRecord indexRecordS)
        {
            if(indexRecordS.VAMin <  indexRecordP.VAMin) 
            {
                indexRecordP.VAMin = indexRecordS.VAMin;
            }
            if(indexRecordS.VBMin <  indexRecordP.VBMin) 
            {
                indexRecordP.VBMin = indexRecordS.VBMin;
            }
            if(indexRecordS.VCMin <  indexRecordP.VCMin) 
            {
                indexRecordP.VCMin = indexRecordS.VCMin;
            }

            if(indexRecordS.VAMax >  indexRecordP.VAMax) 
            {
                indexRecordP.VAMax = indexRecordS.VAMax;
            }
            if(indexRecordS.VBMax >  indexRecordP.VBMax) 
            {
                indexRecordP.VBMax = indexRecordS.VBMax;
            }
            if(indexRecordS.VCMax >  indexRecordP.VCMax) 
            {
                indexRecordP.VCMax = indexRecordS.VCMax;
            }

            // Always a sag event
            if(indexRecordS.Magnitude < indexRecordP.Magnitude) 
            {
                indexRecordP.Magnitude = indexRecordS.Magnitude;
            }
        }
/************************************************************************
* NAME:  FilterEvents
* FUNCTIONAL DESCRIPTION:  Filter events requested by user and delete unwanted events
* INPUT PARAMETERS:
*                     myEventType: TTimeSeriesEventType  event type to filter
*                     myPhase :  THMIEventPhase;   event phase to filter
* OUTPUT PARAMETERS: none
* RETURN: result : none
************************************************************************/
        public void FilterEvents(TTimeSeriesEventType myEventType,
                    THMIEventPhase myPhase)
        {
            TTimeSeriesIndexRecord indexRecord;
            int i,j,count;
            int[] deleteList;  // list of events to delete
            TTimeSeriesRecord dataRecord;
            Boolean foundI,foundO,foundU,foundA;

            deleteList = new int[EventData.IndexTable.Count];  // worst case length
            count = 0;    // number of events to delete

            // Look for events that do not match the event type.

            for( j = 0; j< EventData.IndexTable.Count; j++)
            {
                foundA = false;
                foundI = false;
                foundO = false;
                foundU = false;
                indexRecord = EventData.IndexTable[j];

                for (i = indexRecord.FirstRecord; i<indexRecord.LastRecord; i++)
                {
                    dataRecord = EventData.RawDataTable[i];   // get indexed record from list
                    if(dataRecord.Status == INVALID_RECORD) continue;  // invalid (overflow) record

                    switch (myPhase)
                    {
                        case THMIEventPhase.epA:
                        case THMIEventPhase.epB:
                        case THMIEventPhase.epC:      
                            FindSinglePhaseEvent(myEventType,myPhase,dataRecord,ref foundO,ref foundU,ref foundI,ref foundA);
                        break;

                        case THMIEventPhase.epAll:    // All
                            FindAnyPhaseEvent(myEventType,dataRecord,ref foundO,ref foundU,ref foundI,ref foundA);
                        break;

                        default:
                            System.Diagnostics.Debug.Assert(false, "unknown event phase");
                        break;
                    }
                }

                if(foundA || foundI || foundO || foundU)    // only reset if non-trigger event found
                {
                    ResetUnwantedEvents(myEventType, ref foundO, ref foundU, ref foundI);
                    if(!(foundA || foundI || foundO || foundU)) 
                    {
                        deleteList[count] = j;
                        count++;
                    }
                }
                else if (myEventType != TTimeSeriesEventType.evAll)   // event contains no requested condition. 
                {                               // if filtering event type, delete trigger events
                    deleteList[count] = j;
                    count++;
                }
            }
            DeleteEvents(deleteList,count);
        }
/************************************************************************
* NAME:  DeleteEvents
* FUNCTIONAL DESCRIPTION:  Delete events and renumber them
* INPUT PARAMETERS:   deleteList : array of integer  list of events to delete
*                     count : integer  number of events to delete
* OUTPUT PARAMETERS: none
* RETURN: result : none
************************************************************************/
        public void DeleteEvents(int[] deleteList, int count)
        {
            int i;
            TTimeSeriesIndexRecord indexRecord;

            // delete unwanted events (in reverse order to preserve position)
            for (i = count - 1; i>= 0; i--)
            {
                EventData.IndexTable.Remove(EventData.IndexTable[deleteList[i]]);
            }

            // renumber the events if any were deleted

            if(count > 0) 
            {
                for (i = 0; i<EventData.IndexTable.Count; i++)
                {
                    indexRecord = EventData.IndexTable[i]; // initial sag event
                    indexRecord.EventName = string.Format("#{0:d4}",i+1);
                    indexRecord.Violation = CBEMAViolation(indexRecord); // update violation type
                }
            }
        }
/************************************************************************
* NAME: InitChartTitles
* FUNCTIONAL DESCRIPTION: Set chart titles (main, x and y axes)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************/
        public override void InitChartTitles()
        {
            int i,index, checkCount;
            string title;
            TTimeSeriesIndexRecord dataIndex;
            int adiv,amod;
            TTimeSeriesInputSettingsSSI mySSI;
           
            base.InitChartTitles();
            mySSI = (TTimeSeriesInputSettingsSSI)EventData.InputSettings;

            if(EventData.RawDataTable.Count > 1)    // there is data to be charted
            {
                checkCount = 0;
                title = EventData.MeterID + " Event(s): ";
                index = 0;
                for (i = 0; i<EventData.IndexTable.Count; i++)  // look at event index records
                {
                    dataIndex = EventData.IndexTable[i];
                    if(dataIndex.Checked == false)  continue;  // event not selected for display
                    checkCount++;
                    adiv = System.Math.DivRem(checkCount,10,out amod);
                    title = title + "#" + Convert.ToString(i+1) + ",";
                    if(amod == 0)
                    {
                        UpdateChartTitle(index,title);
                        index++;
                        title = "";
                    }
                }
                if(checkCount > 0)        // some events were checked
                {
                    if(title.Length > 0)    // some remaining events to put in title
                    {
                        UpdateChartTitle(index,title);
                    }
                }
                else
                {
                    if (mySSI.ChartCBEMA) 
                    {
                        ChartDisplay.Header.Lines[0] = EventData.MeterID + "- ITI(CBEMA) Events ";
                    }
                    else
                    {
                        ChartDisplay.Header.Lines[0] = EventData.MeterID + "- Event History";
                    }
                }

                if (mySSI.ChartCBEMA) 
                {
                    ChartDisplay.Axes.Left.Title.Caption = "Amplitude (% of Reference)";
                    ChartDisplay.Axes.Bottom.Title.Caption = "Duration (seconds)";
                }
                else
                {
                    ChartDisplay.Axes.Left.Title.Caption = "Percent of Reference";
                    if(EventData.InputSettings.ChartAxisDateTime)     // display date/time X axis labels
                    {
                        ChartDisplay.Axes.Bottom.Title.Caption = "Time";
                    }
                    else
                    {
                        ChartDisplay.Axes.Bottom.Title.Caption = "Record Number";
                    }
                }

                ChartDisplay.Axes.Left.Title.Angle = 90;
            }
        }
/************************************************************************
* NAME:    UpdateChartTitle
* FUNCTIONAL DESCRIPTION:  update the chart title
* INPUT PARAMETERS: index : int  title position to update
 *                  title : string new title string
* OUTPUT PARAMETERS: 
* RETURN: 
************************************************************************/
        private void UpdateChartTitle(int index, string title)
        {
            title = title.Substring(0,title.Length-1);  // remove trailing comma
            if(ChartDisplay.Header.Lines.Length < index+1) 
            {
                string[] newlines = new string[index+1];
                for (int i = 0; i < ChartDisplay.Header.Lines.Length; i++)  // copy old titles to new list
                {
                    newlines[i] = ChartDisplay.Header.Lines[i];
                }
                newlines[index] = title;  // add new title
                ChartDisplay.Header.Lines = newlines;  // this should free the old list and replace with new list

//                ChartDisplay.Title.Text.Add(title);
            }
            else
            {
                ChartDisplay.Header.Lines[index] = title;
            }
        }
/************************************************************************
* NAME:    AppendCSVHeader
* FUNCTIONAL DESCRIPTION:  Append header records for CSV file to string
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: MeterBuffer : string  file content
* RETURN: result : boolean   status of export (true = success)
************************************************************************/
        public override Boolean AppendCSVHeader(ref string MeterBuffer) 
        {
            int i;      // indices
            TTimeSeriesIndexRecord dataIndex;
            string myDate,myTime, myDuration;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings();

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");
            System.Diagnostics.Debug.Assert(EventData.RawDataLabels != null,"RawDataLabels is null");
            System.Diagnostics.Debug.Assert(EventData.RawDataTable != null,"RawDataTable is null");

            EventData.InitDateTimeFormat(TDateFormatType.ftEnglish_US,ref myFormat);

            if (!base.AppendCSVHeader(ref MeterBuffer)) 
            {
                return false;
            }

            // Table of SSI event records

            if(EventData.IndexTable.Count > 0) 
            {
                MeterBuffer = MeterBuffer + "Event, Date, Time, Duration, Type, Magnitude, Violation, First, Last, Start, End" + CRLF;

                for (i = 0; i<EventData.IndexTable.Count; i++)   // count selected events
                {
                    dataIndex = EventData.IndexTable[i];
                    myDate = gs.DateToStr(dataIndex.DateTime,myFormat);  // use setting date format
                    myTime = string.Format("{0:hh:mm:ss.fff}",dataIndex.DateTime);
                    gs.DurationToTimeString(out myDuration,dataIndex.Duration);
                    MeterBuffer = MeterBuffer +
                                dataIndex.EventName +
                                myDate + 
                                myTime + 
                                myDuration +
                                dataIndex.Description +
                                gs.NonLocaleFloatToStr(dataIndex.Magnitude) + "\"" +
                                CBEMAViolationString(dataIndex.Violation) + "\"" +
                                string.Format("{ 0:d}, {1:d}, {2:d}, {3:d}",      
                                              dataIndex.FirstRecord,
                                              dataIndex.LastRecord,
                                              dataIndex.StartEvent,
                                              dataIndex.StopEvent) + CRLF;
                }
            }

            return true;
        }
/************************************************************************
* NAME:    AppendCSVData
* FUNCTIONAL DESCRIPTION: Append to string a CSV report from HMI data stored
*                         in memory resident TimeSeries structure
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      MeterBuffer : string  file content
* RETURN:                 result : boolean   true if data appended, otherwise false
************************************************************************/
        public override Boolean AppendCSVData(ref string MeterBuffer)
        {
            const string NO_SSI_MSG = "# No SSI data available.";

            int i,j,channelCount,recordCount,pos;      // indices
            string myDate,myTime,myValue;
            TTimeSeriesRecord dataRecord;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings() ;
            double rValue;

            System.Diagnostics.Debug.Assert(EventData != null,"EventData is null");
            System.Diagnostics.Debug.Assert(EventData.RawDataLabels != null,"RawDataLabels is null");
            System.Diagnostics.Debug.Assert(EventData.RawDataTable != null,"RawDataTable is null");
            EventData.InitDateTimeFormat(TDateFormatType.ftEnglish_US, ref myFormat);
            myFormat.ShortDateFormat = "MM/dd/yyyy";

            channelCount = EventData.RawDataLabels.Count;
            recordCount = EventData.RawDataTable.Count;

            if((channelCount == 0) || (recordCount == 0))    // no data available
            {
                MeterBuffer = MeterBuffer + NO_SSI_MSG + CRLF;
                return false;
            }
            else
            {
                MeterBuffer = MeterBuffer + EventData.ChartField + ", Date, Time, ";

                // export channel names
                for (i = 0; i<channelCount; i++) 
                {
                    MeterBuffer = MeterBuffer + EventData.RawDataLabels[i] + ", ";
                }

                MeterBuffer = MeterBuffer + "Phase, Status" + CRLF;

                // export channel data
                for( i = 0 ; i< recordCount; i++)
                {
                    dataRecord = EventData.RawDataTable[i];
                    MeterBuffer = MeterBuffer + string.Format("{0:d}",dataRecord.RecordNumber) + ", ";
                    myDate = gs.DateToStr(dataRecord.DateTime,myFormat);  // use setting date format
                    myTime = string.Format("{0:hh:mm:ss.fff}",dataRecord.DateTime);
                    MeterBuffer = MeterBuffer + myDate + ", " + myTime + ", ";

                    for (j = 0; j<channelCount; j++)
                    {
                        rValue = dataRecord.Value[j] + 0.005;  // account for 2 decimal digit precision error
                        myValue = gs.NonLocaleFloatToStr(rValue);
                        pos = myValue.IndexOf('.');  // look for decimal
                        if(pos > 0) 
                        {
                            myValue = myValue.Substring(0,pos+2);   // keep only 2 decimal places
                        }
                        MeterBuffer = MeterBuffer + myValue + ", ";
                    }

                    MeterBuffer = MeterBuffer + dataRecord.Phase + ", " +
                       dataRecord.Status + CRLF;
                }
                return true;
            }
        }
/************************************************************************
* NAME:    AppendReport
* FUNCTIONAL DESCRIPTION: Append SSI event summary to string
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   MeterBuffer : string  file content
* RETURN:              result : boolean   true if report appended, otherwise false
************************************************************************/
        public override Boolean AppendReport(out string MeterBuffer)
        {
            const string INCOMPLETE_EVENT_MSG = "** = Incomplete Event.";
            const string NO_EVENT_MSG =  "NOTE: No events found.";

            int i,eNumber;
            TTimeSeriesIndexRecord dataIndex;
            string myDate, myTime, myDuration;
            GlobalString gs = new GlobalString();
            TFormatSettings myFormat = gs.InitTFormatSettings();
            string timeSource,eventMag;
            float vbase;
            Boolean hasIncomplete;

            EventData.InitDateTimeFormat(TDateFormatType.ftInternational, ref myFormat);
            gs.DateFormat2DigitYear(ref myFormat);
            hasIncomplete = false;

            vbase = (EventData.SettingTable[0] as TTimeSeriesSettingRecord).SettingValue;
            if(EventData.TimeSource[1] == 'I') 
            {
                timeSource = "int";
            }
            else
            {
                timeSource = "ext";
            }

            if(EventData.IndexTable.Count > 0) 
            {
                myDate = gs.DateToStr(EventData.ChartDateTime,myFormat);  // use setting date format
                myTime = string.Format("{0:hh:mm:ss.fff}",EventData.ChartDateTime);
                MeterBuffer = string.Format("{0,-60}",EventData.MeterID) +
                            " Date: " + myDate + "    Time: " + myTime + CRLF;
                MeterBuffer = MeterBuffer + string.Format("{0,-78}",EventData.MeterTID) + " Time Source: " +
                            timeSource + CRLF + CRLF;
                MeterBuffer = MeterBuffer + "Vbase = " + string.Format("{0:0.00}",vbase) + "V" +
                            string.Format("{0,71}","Voltage (%Vbase)") + CRLF;

                MeterBuffer = MeterBuffer +
                            "   Event    Date        Time      Event Duration   Event        Va            Vb            Vc      ITIC" + CRLF;
                MeterBuffer = MeterBuffer +
                            "#   Type  " + myFormat.ShortDateFormat.ToLower() +
                            "  hh:mm:ss.sss   hhh:mm:ss.sss   Depth    Min    Max    Min    Max    Min    Max  Region" + CRLF + CRLF;
                for (i = 0 ; i<EventData.IndexTable.Count; i++)  // count selected events
                {
                    dataIndex = EventData.IndexTable[i];
                    if (dataIndex.Completion != TTimeSeriesCompleteType.ctComplete) 
                    {
                        hasIncomplete = true;
                    }
                    eNumber = Convert.ToInt32(dataIndex.EventName.Substring(1));  // remove leading character
                    myDate = gs.DateToStr(dataIndex.DateTime,myFormat); // Use setting date format
                    myTime = string.Format("{0:hh:mm:ss.fff}",dataIndex.DateTime);
                    gs.DurationToTimeString(out myDuration,dataIndex.Duration);
                    if(dataIndex.EventType == TTimeSeriesEventType.evNone)      // Trigger event
                    {
                        eventMag = "$$$$$";      // magnitude has no meaning
                    }
                    else
                    {
                        eventMag = string.Format("{0,5:0.0}",dataIndex.Magnitude);
                    }
                    MeterBuffer = MeterBuffer +
                                string.Format("{0:d2}  {1,-5} {2,8}  {3,12}   {4,13}   {5,5}  {6,5:0.0}  {7,5:0.0}  {8,5:0.0}  {9,5:0.0}  {10,5:0.0}  {11,5:0.0}  {12,2}",
                                     eNumber,
                                      dataIndex.Description,
                                      myDate,
                                      myTime,
                                      myDuration,
                                      eventMag,
                                      dataIndex.VAMin,
                                      dataIndex.VAMax,
                                      dataIndex.VBMin,
                                      dataIndex.VBMax,
                                      dataIndex.VCMin,
                                      dataIndex.VCMax,
                                      CBEMAViolationString(dataIndex.Violation)) + CRLF;
                }

                // Only last index entry can contain an incomplete event
                dataIndex = EventData.IndexTable[EventData.IndexTable.Count-1];
                if(hasIncomplete || gs.ContainsText(dataIndex.Description,"*")) 
                {
                    MeterBuffer = MeterBuffer + CRLF + INCOMPLETE_EVENT_MSG;
                }
            }
            else
            {
                MeterBuffer = CRLF + NO_EVENT_MSG;
            }

            return true;
        }
/************************************************************************
* NAME: CBEMAViolation
* FUNCTIONAL DESCRIPTION: return CBEMA violation code based on event duration/magnitude
* INPUT PARAMETERS: indexRecord : TTimeSeriesIndexRecord
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesCBEMAViolationType
************************************************************************/
        public override TTimeSeriesCBEMAViolationType CBEMAViolation( TTimeSeriesIndexRecord indexRecord )
        {
            const int DURATION70 = 20;     // 20ms
            const int DURATION80 = 500;    // 0.5s
            const int DURATION90 = 10000;   // 10s
            const int DURATION110 = 500;   //  0.5s
            const int DURATION120 = 3;     // 3ms
            const int DURATION200 = 1;      // 1ms
            const int MILLISEC_PER_SEC = 1000;
            const int SECONDS_PER_HOUR = 3600;
            const int SECONDS_PER_MINUTE = 60;

            int duration;
            TTimeSeriesCBEMAViolationType violation;
            int hour, min, sec, mSec;
            GlobalString gs = new GlobalString();

            gs.DecodeTime(indexRecord.Duration,out hour,out min,out sec,out mSec);
            duration = (hour * SECONDS_PER_HOUR + min * SECONDS_PER_MINUTE + sec) *
            MILLISEC_PER_SEC + mSec;

            if((indexRecord.Magnitude < 70.0) && (duration > DURATION70)) 
            {
                violation = TTimeSeriesCBEMAViolationType.cvt70;
            }
            else if((indexRecord.Magnitude < 80.0) && (duration > DURATION80)) 
            {
                violation = TTimeSeriesCBEMAViolationType.cvt80;
            }
            else if((indexRecord.Magnitude < 90.0) && (duration > DURATION90)) 
            {
                violation = TTimeSeriesCBEMAViolationType.cvt90;
            }
            else if((indexRecord.Magnitude > 200.0) && (duration > DURATION200)) 
            {
                violation = TTimeSeriesCBEMAViolationType.cvt200;
            }
            else if((indexRecord.Magnitude > 120.0) && (duration > DURATION120)) 
            {
                violation = TTimeSeriesCBEMAViolationType.cvt120;
            }
            else if((indexRecord.Magnitude > 110.0) && (duration > DURATION110)) 
            {
                violation = TTimeSeriesCBEMAViolationType.cvt110;
            }
            else
            {
                violation = TTimeSeriesCBEMAViolationType.cvtNone;
            }
            return violation;
        }
/************************************************************************
* NAME:     CBEMAViolationString
* FUNCTIONAL DESCRIPTION:  Return string describing CBEMA violation
* INPUT PARAMETERS: violation : TTimeSeriesCBEMAViolationType
* OUTPUT PARAMETERS: none
* RETURN:           result : string   violation code
************************************************************************/
        public override string CBEMAViolationString(TTimeSeriesCBEMAViolationType violation)
        {
            string[] VLIST = new string[8]  {"SR", "ND", "ND", "ND", "ND", "PR", "PR", "PR"};

            if(((int)violation >= 0) && ((int)violation < VLIST.Length)) 
            {
                return VLIST[(int)violation];
            }
            else
            {
                System.Diagnostics.Debug.Assert(false,"Invalid violation enumeration");
                return "NA";
            }
        }
        /************************************************************************
       * NAME:    InitActiveChannels
       * FUNCTIONAL DESCRIPTION:  initialize active channel array
       * Save the active series in the chart
        INPUT PARAMETERS:    newCount : int  new number of active channels
       * OUTPUT PARAMETERS:   none
       * RETURN:              none
       ************************************************************************/

        public void InitActiveChannels(int newCount)
        {
            const int VOLTAGE_COUNT = 3;
            int i, currentIndex;

            EventData.InputSettings.SetActiveChannelLength(newCount);
            EventData.InputSettings.ActiveChannels[newCount - 1] = false; // disable last channel (event)

            // position of last current channel
            currentIndex = EventData.RawDataLabels.Count - VOLTAGE_COUNT;

            for (i = 0; i < currentIndex; i++)    // disable the current channels
            {
                EventData.InputSettings.ActiveChannels[i] = false;  // currents
            }

            for (i = currentIndex; i < EventData.RawDataLabels.Count; i++)  // enable voltage channels
            {
                EventData.InputSettings.ActiveChannels[i] = true;   // voltages
            }
        }
    }
}
